/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                           */
/*                  This file is part of the program                         */
/*        SMV2LP -- Converting SMV to Integer Linear Programs                */
/*                                                                           */
/*    Copyright (C) 2002-2005 Tobias Achterberg                              */
/*                            Thorsten Koch                                  */
/*                            Robi Malik                                     */
/*                                                                           */
/*                  2002-2005 Konrad-Zuse-Zentrum                            */
/*                            fuer Informationstechnik Berlin                */
/*                                                                           */
/*  SMV2LP is distributed under the terms of the ZIB Academic License.       */
/*                                                                           */
/*  You should have received a copy of the ZIB Academic License              */
/*  along with SMV2LP; see the file COPYING.                                 */
/*                                                                           */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/* -*-  c-basic-offset: 3 -*- */

#include <stdio.h>
#include <assert.h>
#include <math.h>
#include <limits.h>
#include <string.h>

#include "message.h"
#include "memory.h"
#include "constraint.h"
#include "symbol.h"


/*#define WITHLINES*/
#define CONSLPMAXCOL  8


/****************************************************************************/
/* INDEX - struct ***********************************************************/

struct index
{
   Bool    isatom;      /**< TRUE iff index points to an enumeration atom */
   union
   {
      int        arrayindex;  /**< index of array entry */
      const SYM* atom;        /**< enumeration atom */
   } value;             /**< enumeration atom or array index */
};


/* INDEX - methods **********************************************************/

/**< Creates an array index
 */
static INDEX* consCreateIndexArray(int arrayindex)
{
   INDEX* index;

   allocMemory(&index);
   assert(index != NULL);

   index->isatom = FALSE;
   index->value.arrayindex = arrayindex;

   return index;
}


/**< Creates an enumaration index
 */
static INDEX* consCreateIndexAtom(const SYM* atom)
{
   INDEX* index;

   assert(atom != NULL && atom->type == SYMTYPE_ATOM);

   allocMemory(&index);
   assert(index != NULL);

   index->isatom = TRUE;
   index->value.atom = atom;

   return index;
}


/**< Delete an index and free all associated memory.
 */
static void consDestroyIndex(INDEX* index)
{
   freeMemory(&index);
}


/**< Create a copy of an index.
 *   Variable symbols are shared between the original and the copy.
 */
static INDEX* consCopyIndex(const INDEX* source)
{
   INDEX* dest;

   assert(source != NULL);

   allocMemory(&dest);
   assert(dest != NULL);

   if ( (dest->isatom = source->isatom) ) {
      dest->value.atom = source->value.atom;
   } else {
      dest->value.arrayindex = source->value.arrayindex;
   }

   return dest;
}


/**< Check whether two coefficient indexes are equal.
 *   Two array indexes are equal if they represent the same number.
 *   Two enumeration indexes are equal if they use exactly the same
 *   symbol object.
 */
static Bool consIndexEqual(const INDEX* index1, const INDEX* index2)
{
   if (index1->isatom != index2->isatom) {
      return FALSE;
   } else if (index1->isatom) {
      return index1->value.atom == index2->value.atom;
   } else {
      return index1->value.arrayindex == index2->value.arrayindex;
   }
}


/**< Print a string representation of an index to stdout.
 *   The index is printed as "[<index>]" without trailing newline.
 */
static void consDisplayIndex(const INDEX* index)
{
   if (index == NULL) {
      printf("?NULL?");
   } else if (index->isatom) {
      printf("[%s]", index->value.atom->name);
   } else {
      printf("[%d]", index->value.arrayindex);
   }
}


/****************************************************************************/
/* COEF - struct ************************************************************/

struct indexnode
{
   INDEX*     index;    /**< this index */
   INDEXNODE* next;     /**< rest of index list */
};

struct coef
{
   Longint    factor;    /**< Coefficient of variable in equation. */
   const SYM* varsym;    /**< Symbol of variable. */
   INDEXNODE* headindex; /**< First element of index list. */
   INDEXNODE* tailindex; /**< Last element of index list. */
   TIMESTAT   timestat;  /**< Init, actual, or next timestep. */
};


/* COEF - methods ***********************************************************/

/**< Creates a coefficient.
 */
static COEF* consCreateCoef(Longint factor, const SYM* varsym, TIMESTAT timestat)
{
   COEF* coef;

   allocMemory(&coef);
   assert(coef != NULL);

   coef->factor = factor;
   coef->varsym = varsym;
   coef->headindex = NULL;
   coef->tailindex = NULL;
   coef->timestat = timestat;

   return coef;
}


/**< Delete a coefficient and free all associated memory.
 */
static void consDestroyCoef(COEF* coef)
{
   INDEXNODE* list;
   INDEXNODE* current;

   if (coef != NULL) {
      list = coef->headindex;
      while (list != NULL) {
	 current = list;
	 list = current->next;
	 consDestroyIndex(current->index);
	 freeMemory(&current);
      }
      freeMemory(&coef);
   }
}


/**< Append an index to index list of coefficient.
 */
static void consAppendCoefIndex(COEF* coef, INDEX* index)
{
   INDEXNODE* indexnode;

   assert(coef != NULL);
   assert(index != NULL);

   allocMemory(&indexnode);
   assert(indexnode != NULL);

   indexnode->index = index;
   indexnode->next = NULL;
   if (coef->headindex == NULL) {
      coef->headindex = indexnode;
   } else {
      coef->tailindex->next = indexnode;
   }
   coef->tailindex = indexnode;
}


/**< Create a deep copy of a coefficient.
 */
static COEF* consCopyCoef(const COEF* source)
{
   COEF* dest;
   INDEXNODE* list;
   INDEX* index;

   assert(source != NULL);

   /* Allocate new coefficient and copy simple members ... */
   allocMemory(&dest);
   assert(dest != NULL);

   dest->factor = source->factor;
   dest->varsym = source->varsym;
   dest->timestat = source->timestat;
   dest->headindex = NULL;
   dest->tailindex = NULL;

   /* Copy list of indexes preserving order ... */
   for (list = source->headindex; list != NULL; list = list->next) {
      index = consCopyIndex(list->index);
      consAppendCoefIndex(dest, index);
   }

   return dest;
}


/**< Check whether two coefficients use the same variable.
 *   Two variables are equal, iff
 *   - they use the same symbol object;
 *   - they use the same time status;
 *   - they have equal array indexes.
 */
static Bool consCoefEqual(const COEF* coef1, const COEF* coef2)
{
   INDEXNODE* indexnode1;
   INDEXNODE* indexnode2;

   assert(coef1 != NULL);
   assert(coef2 != NULL);

   if (coef1->varsym != coef2->varsym) {
      return FALSE;
   } else if (coef1->timestat != coef2->timestat) {
      return FALSE;
   } else {
      indexnode1 = coef1->headindex;
      indexnode2 = coef2->headindex;
      while (indexnode1 != NULL) {
	 if (indexnode2 == NULL) {
	    return FALSE;
	 } else if (!consIndexEqual(indexnode1->index, indexnode2->index)) {
	    return FALSE;
	 }
	 indexnode1 = indexnode1->next;
	 indexnode2 = indexnode2->next;
      }
   }

   return TRUE;					 
}


/**< Get a symbol representing the type of this coefficient's variable.
 */
static const SYM* consGetCoefType(const COEF* coef)
{
   static const SYM bintype = {name: NULL, firstvarnum: -1, numvar: 1, type: SYMTYPE_BINVAR};

   const SYM* symtype;
   INDEXNODE* list;

   assert(coef != NULL);
   assert(coef->varsym != NULL);

   symtype = coef->varsym;
   for( list = coef->headindex;	list != NULL; list = list->next )
   {
      switch (symtype->type)
      {
      case SYMTYPE_ARRAY:
	 symtype = symtype->value.array.elemtype;
	 break;
      case SYMTYPE_ENUMVAR:
	 assert(list->next == NULL);  /* Bad index structure! */
	 return &bintype;
      default:
	 assert(FALSE);  /* Bad index structure! */
	 break;
      }
   }
   
   return symtype;
}


/**< Print a string representation of a coefficient to stdout.
 *   If the forcesign flag is set, the leading sign is printed even if
 *   the coefficient's factor is positive.
 *   The coefficient is printed without trailing newline.
 */
static void consDisplayCoef(const COEF* coef, Bool forcesign)
{
   Longint factor;
   INDEXNODE* list;

   if (coef == NULL) {

      printf("?NULL?");

   } else {

      /* Print factor ... */
      factor = coef->factor;
      if (forcesign) {
	 if (factor < 0) {
	    printf(" - ");
	    factor = -factor;
	 } else {
	    printf(" + ");
	 }
      }
      if (factor == 1) {
	 /* don't print */
      } else if (factor == -1) {
	 printf("-");
      } else {
	 printf("%lld * ", factor);
      }

      /* Print symbol name ... */ 
      if (coef->varsym == NULL || coef->varsym->name == NULL) {
	 printf("?NULL?");
      } else {
	 printf("%s", coef->varsym->name);
      }

      /* Print index list ... */
      for (list = coef->headindex; list != NULL; list = list->next) {
	 consDisplayIndex(list->index);
      }

      /* Print time index ... */
      switch (coef->timestat)
      {
      case TIMESTAT_INIT:
	 printf(":[0]");
	 break;
      case TIMESTAT_ACTUAL:
	 printf(":[t]");
	 break;
      case TIMESTAT_NEXT:
	 printf(":[t+1]");
	 break;
      default:
	 printf(":[?%d?]", coef->timestat);
	 break;
      }

   }
}


/**< Outputs the coeficient to file 'f' with LP variable numbers.
 */
static void consCoefToLP(const COEF* coef, int t, int numvar, FILE* f)
{
   INDEXNODE* indexlist;
   INDEX*     index;
   const SYM* varsym;
   int indexoffset;

   assert(coef != NULL);
   assert(f != NULL);

   /* adjust t to actual timestat */
   switch( coef->timestat )
   {
   case TIMESTAT_INIT:
      t = 0;
      break;
   case TIMESTAT_ACTUAL:
      break;
   case TIMESTAT_NEXT:
      t++;
      break;
   default:
      errorMessage("Invalid time status");
      abort();
   }

   /* Calculate variable number */ 
   assert(coef->varsym != NULL && coef->varsym->firstvarnum >= 0);
   indexoffset = 0;
   varsym = coef->varsym;
   for( indexlist = coef->headindex; indexlist != NULL; indexlist = indexlist->next )
   {
      index = indexlist->index;
      if( index->isatom )
      {
         assert(varsym->type == SYMTYPE_ENUMVAR);
         assert(index->value.atom->type == SYMTYPE_ATOM);
         indexoffset += symFindPosition(varsym->value.enumvar.elemlist, index->value.atom->name);
         assert(indexlist->next == NULL);
      }
      else
      {
         assert(varsym->type == SYMTYPE_ARRAY);
         varsym = varsym->value.array.elemtype;
         indexoffset += index->value.arrayindex * symGetNumVariables(varsym);
      }
   }

   /* Output coeficient */
   fprintf(f, " %+lld x%d", coef->factor, t*numvar + coef->varsym->firstvarnum + indexoffset);
}


/* CONSTRAINTFRAGMENT *******************************************************/

/* CONSTRAINTFRAGMENT - struct **********************************************/

struct coeflist
{
   COEF*     coef;      /**< this coefficient */
   COEFLIST* next;      /**< rest of coefficient list */
};


struct constraintfragment
{
   FRAGTYPE  type;        /**< Type of fragment. */
   COEFLIST* coeflist;    /**< List of coefficients.
                           *   The sum of these coefficients represents
			   *   a linear fragment which can be used in
			   *   constraints. Contains only one coefficient
			   *   with factor 1 for enum and array types.
			   */
   Longint   value;       /**< Constant added to sum of coefficients.
			   *   Always 0 for enum and array types.
			   */
   Longint   lower;       /**< lower bound of fragment */
   Longint   upper;       /**< upper bound of fragment */
};


/* CONSTRAINTFRAGMENT - forward *********************************************/

static Bool consIncludeFragmentCoef(CONSTRAINTFRAGMENT* frag, COEF* coef);
static void consInsertFragmentCoef(CONSTRAINTFRAGMENT* frag, COEF* coef);


/* CONSTRAINTFRAGMENT - public **********************************************/

/**< Create a scalar constraint fragment,
 *   representing the constant value 0.
 */
CONSTRAINTFRAGMENT* consCreateFragment(void)
{
   return consCreateFragmentConst(0);
}


/**< Create a scalar constraint fragment,
 *   representing the given constant value.
 */
CONSTRAINTFRAGMENT* consCreateFragmentConst(Longint value)
{
   CONSTRAINTFRAGMENT* frag;

   allocMemory(&frag);
   assert(frag != NULL);

   frag->type = FRAGTYPE_SCALAR;
   frag->coeflist = NULL;
   frag->value = value;
   frag->lower = value;
   frag->upper = value;
   return frag;

}


/**< Create a constraint fragment, representing a single variable.
 *   Depending on the type of the given variable, this method creates
 *   a scalar, enumeration, or array constraint.
 */
CONSTRAINTFRAGMENT* consCreateFragmentVar(const SYM* symbol, TIMESTAT timestat)
{
   CONSTRAINTFRAGMENT* frag;
   COEF* coef;

   assert(symbol != NULL);

   allocMemory(&frag);
   assert(frag != NULL);

   switch (symbol->type)
   {
   case SYMTYPE_ATOM:
   case SYMTYPE_ENUMVAR:
      frag->type = FRAGTYPE_ENUM;
      frag->lower = LONGINT_MIN;
      frag->upper = LONGINT_MAX;
      break;
   case SYMTYPE_ARRAY:
      frag->type = FRAGTYPE_ARRAY;
      frag->lower = LONGINT_MIN;
      frag->upper = LONGINT_MAX;
      break;
   case SYMTYPE_BINVAR:
   case SYMTYPE_INTVAR:
      frag->type = FRAGTYPE_SCALAR;
      frag->lower = symGetMinVal(symbol);
      frag->upper = symGetMaxVal(symbol);
      break;
   default:
      assert(FALSE);  /* Symbol type not used to build constraint fragment! */
      break;
   }

   frag->coeflist = NULL;
   frag->value = 0;
   coef = consCreateCoef(1, symbol, timestat);
   assert(coef != NULL);
   consInsertFragmentCoef(frag, coef);

   return frag;
}


/**< Create a deep copy of a given constrain fragment.
 *   Variable symbols are shared between the original and the copy.
 */
CONSTRAINTFRAGMENT* consCopyFragment(const CONSTRAINTFRAGMENT* source)
{
   CONSTRAINTFRAGMENT* dest;
   COEFLIST* list;
   COEF* coef;
   
   assert(source != NULL);

   allocMemory(&dest);
   assert(dest != NULL);

   dest->type = source->type;
   dest->value = source->value;
   dest->lower = source->lower;
   dest->upper = source->upper;

   dest->coeflist = NULL;

   for (list = source->coeflist; list != NULL; list = list->next) {	
      coef = consCopyCoef(list->coef);
      assert(coef != NULL);
      consInsertFragmentCoef(dest, coef);
   }

   return dest;
}


/**< Delete a constrain fragment and free all associated memory.
 */
void consDestroyFragment(CONSTRAINTFRAGMENT* frag)
{
   COEFLIST* list;
   COEFLIST* current;

   if (frag != NULL) {
      list = frag->coeflist;
      while (list != NULL) {
	 current = list;
	 list = current->next;
	 consDestroyCoef(current->coef);
	 freeMemory(&current);
      }
      freeMemory(&frag);
   }
}


/**< Add a constant value to a constraint fragment.
 *   The constraint fragment must be of scalar type.
 */
void consAddFragmentConst(CONSTRAINTFRAGMENT* frag, Longint value)
{
   assert(frag != NULL);
   assert(frag->type == FRAGTYPE_SCALAR);

   frag->value += value;
   frag->lower += value;
   frag->upper += value;
}


/**< Add a constraint fragment to another.
 *   Both constraint fragments must be of scalar type.
 *   This method adds all components of the given other constraint
 *   fragment to this constraint fragment, destructively modifying
 *   this constraint fragment. The other constraint fragment is
 *   left unmodified.
 */
void consAddFragmentFragment(CONSTRAINTFRAGMENT* frag,
			     const CONSTRAINTFRAGMENT* other)
{
   COEFLIST* list = NULL;
   COEF*     coef = NULL;
   Bool      used;
   Longint   newlower;
   Longint   newupper;

   assert(frag != NULL);
   assert(frag->type == FRAGTYPE_SCALAR);
   assert(other != NULL);
   assert(other->type == FRAGTYPE_SCALAR);
   
   newlower = frag->lower + other->lower;
   newupper = frag->upper + other->upper;

   consAddFragmentConst(frag, other->value);
   for (list = other->coeflist; list != NULL; list = list->next) {
      coef = consCopyCoef(list->coef);
      used = consIncludeFragmentCoef(frag, coef);
      if (!used) {
	 consDestroyCoef(coef);
      }
   }

   frag->lower = newlower;
   frag->upper = newupper;
}


/**< Add an array index to a constraint fragment.
 *   The constraint fragment must be of array type.
 *   This method destructively modifies the contraint fragment, resulting in
 *   a new variable fragment respecting the array's element type.
 */
void consAppendFragmentArrayIndex(CONSTRAINTFRAGMENT* frag, Longint value)
{
   INDEX* index;
   COEF* coef;
   const SYM* symtype;

   assert(frag != NULL);
   assert(frag->type == FRAGTYPE_ARRAY);
   assert(frag->coeflist != NULL);
   assert(frag->coeflist->next == NULL);

   index = consCreateIndexArray(value);
   coef = frag->coeflist->coef;
   consAppendCoefIndex(coef, index);
   symtype = consGetCoefType(coef);

   switch (symtype->type)
   {
   case SYMTYPE_BINVAR:
   case SYMTYPE_INTVAR:
      frag->type = FRAGTYPE_SCALAR;
      frag->lower = symGetMinVal(symtype);
      frag->upper = symGetMaxVal(symtype);
      break;
   case SYMTYPE_ENUMVAR:
      frag->type = FRAGTYPE_ENUM;
      break;
   case SYMTYPE_ARRAY:
      break;
   default:
      assert(FALSE);  /* Unexpected symbol type in array! */
      break;
   }      
}


/**< Add an enumeration index to a constraint fragment.
 *   The constraint fragment must be of enumeration type, and not constant.
 *   This method destructively modifies the contraint fragment, resulting in
 *   a scalar variable fragment representing the indicated enumerarion item.
 */
void consAppendFragmentEnumIndex(CONSTRAINTFRAGMENT* frag, const SYM* atom)
{
   INDEX* index;

   assert(frag != NULL);
   assert(frag->type == FRAGTYPE_ENUM);
   assert(!consIsFragmentConst(frag));
   assert(frag->coeflist != NULL);
   assert(frag->coeflist->next == NULL);
   assert(atom->type == SYMTYPE_ATOM);

   index = consCreateIndexAtom(atom);
   consAppendCoefIndex(frag->coeflist->coef, index);
   frag->type = FRAGTYPE_SCALAR;   
   frag->lower = 0;
   frag->upper = 1;
}


/**< Multiply a constraint fragment by a constant.
 *   The constraint fragment must be of scalar type.
 *   This method multiplies all components of this constraint fragment
 *   by the given constant, destructively modifying this constraint fragment.
 */
void consMultiplyFragment(CONSTRAINTFRAGMENT* frag, Longint factor)
{
   COEFLIST* list = NULL;
   COEFLIST* current =  NULL;
   Longint   newlower;
   Longint   newupper;

   assert(frag != NULL);
   assert(frag->type == FRAGTYPE_SCALAR);

   if (factor != 0) {

      frag->value *= factor;
      for (list = frag->coeflist; list != NULL; list = list->next) {
	 list->coef->factor *= factor;
      }
      if (factor > 0) {
	 frag->lower *= factor;
	 frag->upper *= factor;
      } else {
	 newlower = frag->upper * factor;
	 newupper = frag->lower * factor;
	 frag->lower = newlower;
	 frag->upper = newupper;
      }

   } else {

      /* Multiply by 0? --- Delete coefficients and set value to 0 ... */
      frag->value = 0;
      list = frag->coeflist;
      while (list != NULL) {
	 current = list;
	 list = current->next;
	 consDestroyCoef(current->coef);
	 freeMemory(&current);
      }
      frag->coeflist = NULL;
      frag->lower = 0;
      frag->upper = 0;

   }      
}


/**< Negate a constraint fragment.
 *   This method destructively replaces the constraint fragment <frag>
 *   by a fragment representing 1 - <frag>.
 *   The constraint fragment <frag> must be of binary scalar type.
 */
void consNegateFragment(CONSTRAINTFRAGMENT* frag)
{
   assert(frag != NULL);
   assert(frag->type == FRAGTYPE_SCALAR);
   assert(consGetFragmentMinVal(frag) >= 0);
   assert(consGetFragmentMaxVal(frag) <= 1);

   consMultiplyFragment(frag, -1);
   consAddFragmentConst(frag, 1);
}


/**< Get the type of this constraint fragment.
 */
FRAGTYPE consGetFragmentType(const CONSTRAINTFRAGMENT* frag)
{
   assert(frag != NULL);
   return frag->type;
}


/**< Get a symbol representing the type of this constraint fragment.
 *   The fragment must be represented by a single variable or enumeration
 *   constant.
 */
const SYM* consGetFragmentTypeSym(const CONSTRAINTFRAGMENT* frag)
{
   assert(frag != NULL);
   assert(frag->coeflist != NULL);
   assert(frag->coeflist->next == NULL);

   return consGetCoefType(frag->coeflist->coef);
}


/**< Check whether this constraint fragment represents a symbol,
 *   that means, if there is only a single coeficient with value 1,
 *   and the additive constant is 0.
 */
Bool consIsFragmentSymbol(const CONSTRAINTFRAGMENT* frag)
{
   assert(frag != NULL);

   return( frag->coeflist != NULL && frag->coeflist->next == NULL
      && frag->coeflist->coef->factor == 1 && frag->value == 0 );
}

/**< Get the symbol of an enumeration or array constraint fragment.
 */
const SYM* consGetFragmentSymbol(const CONSTRAINTFRAGMENT* frag)
{
   assert(consIsFragmentSymbol(frag));

   return frag->coeflist->coef->varsym;
}


/**< Check whether this constraint fragment represents a constant.
 *   A scalar fragment is constant if it has no coefficients.
 *   A enumeration fragment is constant if its symbol is an atom.
 *   An array fragment is never constant.
 */
Bool consIsFragmentConst(const CONSTRAINTFRAGMENT* frag)
{
   assert(frag != NULL);

   switch (frag->type)
   {
   case FRAGTYPE_SCALAR:
      return frag->coeflist == NULL;
   case FRAGTYPE_ENUM:
      return consGetFragmentSymbol(frag)->type == SYMTYPE_ATOM;
   case FRAGTYPE_ARRAY:
      return FALSE;
   default:
      assert(FALSE);  /* Unknown fragment type! */
      return FALSE;
   }
}


/**< Returns TRUE iff the given scalar constraint fragment represents a
 *   binary value.
 */
Bool consIsFragmentBoolean(const CONSTRAINTFRAGMENT* frag)
{
   assert(frag != NULL);
   assert(consGetFragmentType(frag) == FRAGTYPE_SCALAR);

   return(consGetFragmentMinVal(frag) >= 0 && consGetFragmentMaxVal(frag) <= 1);
}


/**< Get the minimum possible value of a constraint fragment.
 *   The constraint fragment must be of scalar type.
 */
Longint consGetFragmentMinVal(const CONSTRAINTFRAGMENT* frag)
{
   assert(frag != NULL);
   assert(frag->type == FRAGTYPE_SCALAR);

   return frag->lower;
}


/**< Get the maximum possible value of a constraint fragment.
 *   The constraint fragment must be of scalar type.
 */
Longint consGetFragmentMaxVal(const CONSTRAINTFRAGMENT* frag)
{
   assert(frag != NULL);
   assert(frag->type == FRAGTYPE_SCALAR);

   return frag->upper;
}


/**< Sets the bounds of the fragment explicitly.
 */
void consSetFragmentBounds(CONSTRAINTFRAGMENT* frag, Longint lower, Longint upper)
{
   assert(frag != NULL);
   assert(frag->type == FRAGTYPE_SCALAR);

   frag->lower = lower;
   frag->upper = upper;
}


static int getlog2(Longint arg)
{
   assert(arg > 0);
   return (Longint)(ceil(log((double)arg)/log(2.0)));
}


/**< Get the number of bits necessary to store the given scalar type fragment.
 *   Note, that for possible negative values, the corresponding sign bit is
 *   included in the bitsize.
 *   bit decomposition:
 *      non-negative:   x = 2^0 b0 + 2^1 b1 + ... + 2^(n-1) b(n-1)
 *      negative:       x = 2^0 b0 + 2^1 b1 + ... + 2^(n-2) b(n-2) - 2^(n-1) b(n-1)
 */
int consGetFragmentBitlength(const CONSTRAINTFRAGMENT* frag)
{
   Longint minval;
   Longint maxval;
   int bitlength = 0;

   assert(frag != NULL);
   assert(frag->type == FRAGTYPE_SCALAR);

   minval = consGetFragmentMinVal(frag);
   maxval = consGetFragmentMaxVal(frag);

   /* maxval > 0: 2^(bitlength)-1 >= maxval */
   if( maxval > 0 )
      bitlength = getlog2(maxval+1);
   /* minval < 0: 2^(bitlength-1) >= -minval */
   if( minval < 0 )
      bitlength = MAX(bitlength, getlog2(-minval)) + 1;

   return bitlength;
}


/**< Print a string representation of a constraint fragment to stdout.
 *   The constraint fragment is printed in human-readable form on a line
 *   of its own.
 */
void consDisplayFragment(const CONSTRAINTFRAGMENT* frag)
{
   COEFLIST* list;
   Bool forcesign;

   if (frag == NULL)
   {
      printf("?NULL?\n");
   }
   else
   {
      /* Print sum of coefficients ... */
      if (frag->coeflist == NULL) {
	 printf("0");
      } else {
	 forcesign = FALSE;
	 for (list = frag->coeflist; list != NULL; list = list->next) {
	    consDisplayCoef(list->coef, forcesign);
	    forcesign = TRUE;
	 }
      }
      printf("\n");
   }
}


/* CONSTRAINTFRAGMENT - private *********************************************/

/**< Add coefficient to this constraint fragment.
 *   The constraint fragment must be of scalar type.
 *   This method checks whether the constraint fragment already contains
 *   a coefficient using the same variable, and if so, it simply adds the two
 *   factors. If not, the given coefficient is added to the list.
 *   Note that this method does not update the bounds of the fragment.
 *   @returns TRUE if the given coefficient object is now included in the
 *            constraint fragment. Otherwise it has been added to an existing
 *            coefficient, and the caller should consider deleting it.
 */
static Bool consIncludeFragmentCoef(CONSTRAINTFRAGMENT* frag, COEF* coef)
{
   COEFLIST** listref = NULL;
   COEFLIST* list = NULL;
   COEF* listcoef = NULL;

   assert(frag != NULL);
   assert(frag->type == FRAGTYPE_SCALAR);
   assert(coef != NULL);
   
   /* Search coefficient for existing variable ... */
   for (listref = &frag->coeflist;
	*listref != NULL;
	listref = &(*listref)->next) {
      listcoef = (*listref)->coef;
      if (consCoefEqual(listcoef, coef)) {
	 /* Variable found: add factors ... */
	 listcoef->factor += coef->factor;
	 if (listcoef->factor == 0) {
	    /* Sum of factors is zero: delete coefficient from list ... */
	    consDestroyCoef(listcoef);
	    list = *listref;
	    *listref = list->next;
	    freeMemory(&list);
	 }
	 return FALSE;
      }
   }

   /* Variable not found: include new coefficient ... */
   consInsertFragmentCoef(frag, coef);

   return TRUE;
}


/**< Add a coefficient to this constraint fragment.
 *   This method prepends the given coefficient to the coefficient list
 *   of this constraint fragment. No type check is performed.
 */
static void consInsertFragmentCoef(CONSTRAINTFRAGMENT* frag, COEF* coef)
{
   COEFLIST* list = NULL;

   assert(frag != NULL);
   assert(coef != NULL);

   allocMemory(&list);
   assert(list != NULL);
   list->coef = coef;
   list->next = frag->coeflist;
   frag->coeflist = list;
}


/****************************************************************************/
/* CONSTRAINT - struct ******************************************************/

struct constraint
{
   COEFLIST* coeflist;
   INEQTYPE  ineqtype;
   Longint   rhs;
   char*     comment;
};


/* CONSTRAINT - public ******************************************************/

/**< Creates a constraint.
 */
CONSTRAINT* consCreate(INEQTYPE ineqtype, Longint rhs)
{
   CONSTRAINT* cons;

   allocMemory(&cons);
   assert(cons != NULL);

   cons->coeflist = NULL;
   cons->ineqtype = ineqtype;
   cons->rhs = rhs;
   cons->comment = NULL;

   return cons;
} 

/**< Create a constraint from a given constraint fragment.
 *   This method produces a constraint asserting that a given constraint
 *   fragment is equal to/greater or equal to/less or equal to the
 *   given right hand side.
 *   The method may also return NULL if the constraint is found to be 
 *   trivially true after inspecting the fragment's bounds.
 *   @returns Created constraint or NULL.
 */
CONSTRAINT* consCreateFromFragment(
   const CONSTRAINTFRAGMENT* frag,
   INEQTYPE ineqtype,
   Longint rhs)
{
   CONSTRAINT* cons = NULL;
   COEFLIST* list = NULL;
   COEF* coef = NULL;
   Longint normfactor = 1;

   assert(frag != NULL);

   /* Check if constraint is trivially true ... */
   switch (ineqtype)
   {
   case INEQTYPE_LESS:
      if (consGetFragmentMaxVal(frag) <= rhs) {
	 return NULL;
      }
      break;
   case INEQTYPE_GREATER:
      if (consGetFragmentMinVal(frag) >= rhs) {
	 return NULL;
      }
      normfactor = -1;
      ineqtype = INEQTYPE_LESS;
      break;
   case INEQTYPE_EQUAL:
      if (consGetFragmentMinVal(frag) == rhs &&
	  consGetFragmentMaxVal(frag) == rhs) {
	 return NULL;
      }
      break;
   default:
      printf("bad inequality type %d\n", ineqtype);
      abort();
   }

   /* Create constraint object ... */
   cons = consCreate(ineqtype, normfactor * (rhs - frag->value));
   assert(cons != NULL);

   /* Copy coefficients ... */
   for (list = frag->coeflist; list != NULL; list = list->next) {	
      coef = consCopyCoef(list->coef);
      coef->factor *= normfactor;
      assert(coef != NULL);
      consAddCoef(cons, coef);
   }

   return cons;
}

/**< creates the special property constraint from a given constraint fragment
 *   @returns Created constraint
 */
CONSTRAINT* consCreateProperty(
   const CONSTRAINTFRAGMENT* frag)
{
   CONSTRAINT* cons = NULL;
   COEF* coef = NULL;

   assert(frag != NULL);
   assert(frag->value == 0 || frag->value == 1);

   /* Create constraint object ... */
   cons = consCreate(INEQTYPE_PROPERTY, -frag->value);
   assert(cons != NULL);

   /* fragment may be 0, 1, x_i, or 1-x_i */
   if( frag->coeflist != NULL )
   {
      assert(frag->coeflist->next == NULL); /* property must consist of a single variable */
      assert(frag->coeflist->coef != NULL);
      assert((frag->coeflist->coef->factor == 1 && frag->value == 0)
         || (frag->coeflist->coef->factor == -1 && frag->value == 1)); /* coefficient must be either x_i or 1-x_i */

      /* Copy single coefficient */
      coef = consCopyCoef(frag->coeflist->coef);
      assert(coef != NULL);
      consAddCoef(cons, coef);
   }

   return cons;
}

/**< Creates a comment constraint
 */
CONSTRAINT* consCreateComment(const char* comment)
{
   CONSTRAINT* cons;

   allocMemory(&cons);
   assert(cons != NULL);

   cons->coeflist = NULL;
   cons->ineqtype = INEQTYPE_COMMENT;
   cons->rhs = 0;
   duplicateMemoryArray(&cons->comment, comment, strlen(comment)+1);

   return cons;
}

/**< Creates a comment constraint for an operation
 */
CONSTRAINT* consCreateOpComment(
   const char* opname, 
   const SYM* symres,
   const CONSTRAINTFRAGMENT* fragop1,
   const CONSTRAINTFRAGMENT* fragop2,
   const CONSTRAINTFRAGMENT* fragop3)
{
   char resname[1024] = "";
   char op1name[1024] = "";
   char op2name[1024] = "";
   char op3name[1024] = "";
   char comment[1024];
   
   if( symres != NULL )
      sprintf(resname, "<%s> = ", symGetName(symres));

   if( fragop1 != NULL )
   {
      if( consIsFragmentSymbol(fragop1) )
         sprintf(op1name, "<%s>", symGetName(consGetFragmentSymbol(fragop1)));
      else if( consIsFragmentConst(fragop1) )
         sprintf(op1name, "%lld", consGetFragmentMinVal(fragop1));
      else
         sprintf(op1name, "[%p]", fragop1);
   }

   if( fragop2 != NULL )
   {
      if( consIsFragmentSymbol(fragop2) )
         sprintf(op2name, ",<%s>", symGetName(consGetFragmentSymbol(fragop2)));
      else if( consIsFragmentConst(fragop2) )
         sprintf(op2name, ",%lld", consGetFragmentMinVal(fragop2));
      else
         sprintf(op2name, ",[%p]", fragop2);
   }

   if( fragop3 != NULL )
   {
      if( consIsFragmentSymbol(fragop3) )
         sprintf(op3name, ",<%s>", symGetName(consGetFragmentSymbol(fragop3)));
      else if( consIsFragmentConst(fragop3) )
         sprintf(op3name, ",%lld", consGetFragmentMinVal(fragop3));
      else
         sprintf(op3name, ",[%p]", fragop3);
   }

   sprintf(comment, "  %s%s(%s%s%s)", resname, opname, op1name, op2name, op3name);

   return consCreateComment(comment);
}

#if WITHLINES
static int lastlinecomment = -1;
#endif

/**< Creates a comment constraint for a line number
 */
CONSTRAINT* consCreateLineComment(
   const Node* node)
{
#if WITHLINES
   char comment[1024];

   if( node == NULL || node->lineno == lastlinecomment || node->lineno == -1 )
      return NULL;

   lastlinecomment = node->lineno;

   sprintf(comment, "(line %d)", node->lineno);

   return consCreateComment(comment);
#else
   return NULL;
#endif
}

/**< Adds coefficient to constraint.
 */
void consAddCoef(CONSTRAINT* cons, COEF* coef)
{
   COEFLIST* coeflist;

   assert(cons != NULL);
   assert(coef != NULL);
   
   allocMemory(&coeflist);
   assert(coeflist != NULL);

   coeflist->coef = coef;
   coeflist->next = cons->coeflist;
   cons->coeflist = coeflist;
}


/**< Print a string representation of a constraint to stdout.
 *   The constraint is printed in human-readable form on a line of its own.
 */
void consDisplay(const CONSTRAINT* cons)
{
   COEFLIST* list;
   Bool forcesign;

   if (cons == NULL)
      printf("?NULL?\n");
   else if( cons->ineqtype == INEQTYPE_COMMENT )
   {
      assert(cons->comment != NULL);
      printf(cons->comment);
      printf("\n");
   }
   else if( cons->ineqtype == INEQTYPE_PROPERTY )
   {
      assert(cons->coeflist != NULL);
      assert(cons->coeflist->next == NULL);
      printf("(property) for all t: ");
      consDisplayCoef(cons->coeflist->coef, FALSE);
      printf("\n");
   }
   else
   {
      /* Print sum of coefficients ... */
      if (cons->coeflist == NULL) {
	 printf("0");
      } else {
	 forcesign = FALSE;
	 for (list = cons->coeflist; list != NULL; list = list->next) {
	    consDisplayCoef(list->coef, forcesign);
	    forcesign = TRUE;
	 }
      }

      /* Print inequality sign */
      switch (cons->ineqtype)
      {
      case INEQTYPE_LESS:
	 printf(" <= ");
	 break;
      case INEQTYPE_EQUAL:
	 printf(" = ");
	 break;
      case INEQTYPE_GREATER:
	 printf(" >= ");
	 break;
      default:
	 printf("?%d?", cons->ineqtype);
	 break;
      }

      /* Print right-hand side */
      printf("%lld\n", cons->rhs);
   }
}


/**< Outputs constraint into file 'f' in CPLEX LP format.
 */
static void consConsToLP(const CONSTRAINT* cons, int maxT, int numvar, FILE* f)
{
   assert(f != NULL);
   assert(cons != NULL);
   assert(maxT >= 0);

   if( cons->ineqtype == INEQTYPE_COMMENT )
   {
      assert(cons->comment != NULL);
      fprintf(f, "\\");
      fprintf(f, cons->comment);
      fprintf(f, "\n");
   }
   else if( cons->ineqtype == INEQTYPE_PROPERTY )
   {
      int column;
      int t;

      assert(cons->rhs == 0 || cons->rhs == -1);

      if( cons->coeflist != NULL )
      {
         assert(cons->coeflist->next == NULL); /* property must consist of exactly one variable */
         assert(cons->coeflist->coef != NULL);
         assert((cons->coeflist->coef->factor == 1 && cons->rhs == 0)
            || (cons->coeflist->coef->factor == -1 && cons->rhs == -1)); /* coefficient must be either x_i or 1-x_i */

         /* create property: sum x_t <= |T|-1 (= maxT) */
         column = 0;
         for( t = 0; t <= maxT; ++t )
         {
            consCoefToLP(cons->coeflist->coef, t, numvar, f);
            column = (column+1) % CONSLPMAXCOL;
            if( column == 0 )
               fprintf(f, "\n");
         }
      }
      fprintf(f, " <= %lld\n", (Longint)maxT + (maxT+1)*cons->rhs);
   }
   else
   {
      COEFLIST* coeflist;
      int column;
      int numT;
      int t;
      Bool timeNext = FALSE;
      Bool timeActual = FALSE;

      /* Evaluate timestats of coefficients */
      for( coeflist = cons->coeflist; coeflist != NULL; coeflist = coeflist->next )
      {
         switch( coeflist->coef->timestat )
         {
         case TIMESTAT_INIT:
            break;
         case TIMESTAT_ACTUAL:
            timeActual = TRUE;
            break;
         case TIMESTAT_NEXT:
            timeNext = TRUE;
            break;
         default:
            errorMessage("Invalid time status");
            abort();
         }
      }
      if( timeNext )
         numT = maxT;        /* create constraints for  t = 0, ..., maxT-2 */
      else if( timeActual )
         numT = maxT+1;      /* create constraints for  t = 0, ..., maxT-1 */
      else
         numT = 1;           /* create constraints for  t = 0 */

      for( t = 0; t < numT; ++t )
      {
         /* Print sum of coefficients ... */
         if( cons->coeflist != NULL )
         {
            column = 0;
            for( coeflist = cons->coeflist; coeflist != NULL; coeflist = coeflist->next )
            {
               consCoefToLP(coeflist->coef, t, numvar, f);
               column = (column+1) % CONSLPMAXCOL;
               if( column == 0 )
                  fprintf(f, "\n");
            }
         }
      
         /* Print inequality sign */
         switch( cons->ineqtype )
         {
         case INEQTYPE_LESS:
            fprintf(f, " <= ");
            break;
         case INEQTYPE_EQUAL:
            fprintf(f, " = ");
            break;
         case INEQTYPE_GREATER:
            fprintf(f, " >= ");
            break;
         default:
            errorMessage("Invalid inequality type");
            abort();
         }
      
         /* Print right-hand side */
         fprintf(f, "%lld\n", cons->rhs);
      }
   }
}
   
/****************************************************************************/
/* CONSTRAINTSET: struct ****************************************************/

struct constraintlist
{
   CONSTRAINT*     cons;  /**< this constraint */
   CONSTRAINTLIST* next;  /**< rest of constraint list */
};

struct constraintset
{
   SYMTABLE*        symtable;  /**< symbol table used for variables and enumeration values */
   int              nextsym;   /**< number given to next auxiliary symbol */
   CONSTRAINTLIST*  clist;     /**< list of constraints */
   CONSTRAINTLIST** clistptr;  /**< pointer to last listpointer (pointing to NULL) */
};


/* CONSTRAINTSET: public ****************************************************/

/**< Creates an empty constraint set.
 */
CONSTRAINTSET* consCreateCSet(SYMTABLE* symtable)
{
   CONSTRAINTSET* cset;

   allocMemory(&cset);
   assert(cset != NULL);

   cset->symtable = symtable;
   cset->nextsym = 1;
   cset->clist = NULL;
   cset->clistptr = &(cset->clist);

   return cset;
}


/**< Print a constraint set to stdout.
 *   First the list of variables is printed, followed by the list of
 *   constraints.
 */
void consDisplayCSet(const CONSTRAINTSET* cset)
{
   CONSTRAINTLIST* list;

   if (cset == NULL)
      printf("?cset == NULL?\n");
   else
   {
      /* Print list of variables ... */
      printf("\nVARIABLES:\n\n");
      symDisplayTable(cset->symtable);

      /* Print list of constraints ... */
      printf("\nCONSTRAINTS:\n\n");
      for( list = cset->clist; list != NULL; list = list->next )
	 consDisplay(list->cons);
   }
}


/**< Outputs constraint set to file in CPLEX LP format with time horizon 'maxT'.
 */
void consCSetToLP(const CONSTRAINTSET* cset, int maxT, FILE* f)
{
   CONSTRAINTLIST* list = NULL;
   SYMTABLE*       symtable = NULL;
   int             numvar;
   int             i;
   int             column;

   assert(maxT >= 0);
   assert(f != NULL);

   /* get number of LP variables */
   numvar = consGetNumVariables(cset);
   if( numvar <= 0 )
   {
      printf("LP is empty. No variables.\n");
      return;
   }

   /* objective function: min 1^T x */
   fprintf(f, "min");
   column = 0;
   for( i = 0; i < (maxT+1) * numvar; ++i )
   {
      fprintf(f, " + x%d", i);
      column = (column+1) % CONSLPMAXCOL;
      if( column == 0 )
         printf("\n   ");
   }

   /* constraints */
   fprintf(f, "\nsuch that\n");
   for( list = cset->clist; list != NULL; list = list->next )
      consConsToLP(list->cons, maxT, numvar, f);
   
   /* bounds */
   fprintf(f, "\nBounds\n");
   for ( symtable = cset->symtable; symtable != NULL; symtable = symtable->next )
      symBoundsToLP(symtable->sym, maxT, numvar, f);

   /* integrality section */
   fprintf(f, "\nGeneral\n");
   column = 0;
   for( i = 0; i < (maxT+1) * numvar; ++i )
   {
      fprintf(f, " x%d", i);
      column = (column+1) % CONSLPMAXCOL;
      if( column == 0 )
         printf("\n");
   }

   /* end */
   fprintf(f, "\n\nEnd\n");
}



/**< Adds constraint to constraint set.
 *   This method checks if the given constraint is not NULL,
 *   and if so adds it to this constraint set.
 */
void consAddToSet(CONSTRAINTSET* cset, CONSTRAINT* cons)
{
   CONSTRAINTLIST* newclist = NULL;

   assert(cset != NULL);
   assert(cset->clistptr != NULL);

   if (cons != NULL)
   {
#ifdef DEBUG
      consDisplay(cons);
#endif /* DEBUG */

      allocMemory(&newclist);
      assert(newclist != NULL);

      newclist->cons = cons;
      newclist->next = NULL;
      *(cset->clistptr) = newclist;
      cset->clistptr = &(newclist->next);
   }
}


/**< Creates a new auxiliary binary symbol and adds it to the symbol list of
 *   this constraint set.
 *   @returns The created symbol.
 */
SYM* consCreateBinarySymbol(CONSTRAINTSET* cset)
{
   char name[12];
   SYM* sym;
   int restat;

   assert(cset != NULL);

   sprintf(name, "$%d", cset->nextsym++);
   sym = symCreateBinvar(name);
   assert(sym != NULL);
   restat = symAdd(&cset->symtable, sym);
   assert(restat == SUCCESS);

   return sym;
}


/**< Creates a new auxiliary integer symbol with given lower and upper
 *   bounds and adds it to the symbol list of this constraint set.
 *   @returns The created symbol.
 */
SYM* consCreateIntSymbol(CONSTRAINTSET* cset, Longint lower, Longint upper)
{
   char name[12];
   SYM* sym;
   int restat;

   assert(cset != NULL);

   if (lower >= 0 && upper <= 1) {
      /* Really is a binary symbol ... */
      sym = consCreateBinarySymbol(cset);
   } else {
      sprintf(name, "$%d", cset->nextsym++);
      sym = symCreateIntvar(name, lower, upper);
      assert(sym != NULL);
      restat = symAdd(&cset->symtable, sym);
      assert(restat == SUCCESS);
   }

   return sym;
}


/**< Creates a new auxiliary enumeration symbol and add it to the symbol list
 *   of this constraint set. No enumeration entries are created---use
 *   symAddEnumEntry() to add them.
 *   @returns The created symbol.
 */
SYM* consCreateEnumSymbol(CONSTRAINTSET* cset)
{
   char name[12];
   SYM* sym;
   int restat;

   assert(cset != NULL);

   sprintf(name, "$%d", cset->nextsym++);
   sym = symCreateEnumvar(name);
   assert(sym != NULL);
   restat = symAdd(&cset->symtable, sym);
   assert(restat == SUCCESS);

   return sym;
}


/**< Creates a new auxiliary 'array[first..last] of boolean' symbol and add it to the symbol list
 *   of this constraint set. 
 *   @returns The created symbol.
 */
SYM* consCreateBoolArraySymbol(CONSTRAINTSET* cset, int first, int last)
{
   char name[12];
   SYM* sym;
   int restat;

   assert(cset != NULL);

   sprintf(name, "$%d", cset->nextsym++);
   sym = symCreateArray(name, first, last, symCreateBinvar(NULL));
   assert(sym != NULL);
   restat = symAdd(&cset->symtable, sym);
   assert(restat == SUCCESS);

   return sym;
}


/**< Creates a new auxiliary 'array[first..last] of lower..upper' symbol and add it to the symbol list
 *   of this constraint set. 
 *   @returns The created symbol.
 */
SYM* consCreateIntArraySymbol(CONSTRAINTSET* cset, int first, int last, Longint lower, Longint upper)
{
   char name[12];
   SYM* sym;
   int restat;

   assert(cset != NULL);

   sprintf(name, "$%d", cset->nextsym++);
   sym = symCreateArray(name, first, last, symCreateIntvar(NULL, lower, upper));
   assert(sym != NULL);
   restat = symAdd(&cset->symtable, sym);
   assert(restat == SUCCESS);

   return sym;
}


SYM* consFindSymbol(CONSTRAINTSET* cset, const char* name)
{
   assert(cset != NULL);
   return symFind(cset->symtable, name);
}


/**< Creates equalities for enums:  x_atom1 + x_atom2 + ... + x_atomN = 1
 */
void consCreateEnumEqualities(CONSTRAINTSET* cset)
{
   SYMTABLE*           symtable;
   SYM*                sym;
   SYMTABLE*           elemlist;
   CONSTRAINTFRAGMENT* frag;
   CONSTRAINTFRAGMENT* atomfrag;
   Bool                commentwritten = FALSE;

   assert(cset != NULL);

   symtable = cset->symtable;
   while( symtable != NULL )
   {
      sym = symtable->sym;
      if( sym->type == SYMTYPE_ENUMVAR )
      {
         elemlist = sym->value.enumvar.elemlist;
         frag = consCreateFragment();
         while( elemlist != NULL )
         {
            atomfrag = consCreateFragmentVar(sym, TIMESTAT_ACTUAL);
            consAppendFragmentEnumIndex(atomfrag, elemlist->sym);
            consAddFragmentFragment(frag, atomfrag);
            consDestroyFragment(atomfrag);
            elemlist = elemlist->next;
         }
         if( !commentwritten )
         {
            consAddToSet(cset, consCreateComment("ENUM equalities"));
            commentwritten = TRUE;
         }
         consAddToSet(cset, consCreateFromFragment(frag, INEQTYPE_EQUAL, 1) );
         consDestroyFragment(frag);
      }
      symtable = symtable->next;
   }
}


/**< Counts the number of constraints in the constraint set.
 */
int consGetNumConstraints(const CONSTRAINTSET* cset)
{
   CONSTRAINTLIST* clist;
   int num = 0;

   assert(cset != NULL);
   clist = cset->clist;
   while( clist != NULL )
   {
      if( clist->cons->ineqtype != INEQTYPE_COMMENT )
         num++;
      clist = clist->next;
   }

   return num;
}


/**< Counts the number of variables in the MIP formulation of the constraint set.
 */
int consGetNumVariables(const CONSTRAINTSET* cset)
{
   SYMTABLE* symtable;
   int       num = 0; 

   assert(cset != NULL);

   symtable = cset->symtable;
   while( symtable != NULL )
   {
      num += symGetNumVariables(symtable->sym);
      symtable = symtable->next;
   }

   return num;
}


/**< Assigns MIP variable numbers for the symbols in the constraint set.
 *   @returns number of MIP variables used
 */
int consAssignVariableNumbers(CONSTRAINTSET* cset)
{
   SYMTABLE* symtable;
   int       num = 0; 

   assert(cset != NULL);

   symtable = cset->symtable;
   while( symtable != NULL )
   {
      symSetVariableNumber(symtable->sym, num);
      num += symGetNumVariables(symtable->sym);
      symtable = symtable->next;
   }

   return num;   
}

