/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                           */
/*                  This file is part of the program                         */
/*        SMV2LP -- Converting SMV to Integer Linear Programs                */
/*                                                                           */
/*    Copyright (C) 2002-2005 Tobias Achterberg                              */
/*                            Thorsten Koch                                  */
/*                            Robi Malik                                     */
/*                                                                           */
/*                  2002-2005 Konrad-Zuse-Zentrum                            */
/*                            fuer Informationstechnik Berlin                */
/*                                                                           */
/*  SMV2LP is distributed under the terms of the ZIB Academic License.       */
/*                                                                           */
/*  You should have received a copy of the ZIB Academic License              */
/*  along with SMV2LP; see the file COPYING.                                 */
/*                                                                           */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/* -*-  c-basic-offset: 3 -*- */

#ifndef __CONSTRAINT_H__
#define __CONSTRAINT_H__

#include "def.h"
#include "symbol.h"


enum timestat
{
   TIMESTAT_INIT   = 0, /**< initial time step t = 0 */
   TIMESTAT_ACTUAL = 1, /**< actual time step t */
   TIMESTAT_NEXT   = 2  /**< next time step t+1 */
};
typedef enum timestat TIMESTAT;

enum ineqtype
{
   INEQTYPE_LESS     = 0, /**<rowvalue <= rhs */
   INEQTYPE_EQUAL    = 1, /**< rowvalue == rhs */
   INEQTYPE_GREATER  = 2, /**< rowvalue >= rhs */
   INEQTYPE_PROPERTY = 3, /**< property constraint: \sum x_t <= |T|-1 */
   INEQTYPE_COMMENT  = 4  /**< comment stored in comment string */
};
typedef enum ineqtype INEQTYPE;

enum fragtype
{
   FRAGTYPE_SCALAR  = 0, /**< integer or binary fragment */
   FRAGTYPE_ENUM    = 1, /**< enumeration variable or constant */
   FRAGTYPE_ARRAY   = 2  /**< array not yet fully indexed */
};
typedef enum fragtype FRAGTYPE;


typedef struct index INDEX;

typedef struct indexnode INDEXNODE;

typedef struct coef COEF;

typedef struct coeflist COEFLIST;

typedef struct constraintfragment CONSTRAINTFRAGMENT;

typedef struct constraint CONSTRAINT;

typedef struct constraintlist CONSTRAINTLIST;

typedef struct constraintset CONSTRAINTSET;


/* CONSTRAINTFRAGMENT */
extern CONSTRAINTFRAGMENT* consCreateFragment(void);
extern CONSTRAINTFRAGMENT* consCreateFragmentConst(Longint value);
extern CONSTRAINTFRAGMENT* consCreateFragmentVar(const SYM* symbol, TIMESTAT timestat);
extern CONSTRAINTFRAGMENT* consCopyFragment(const CONSTRAINTFRAGMENT* source);
extern void consDestroyFragment(CONSTRAINTFRAGMENT* frag);
extern void consAddFragmentConst(CONSTRAINTFRAGMENT* frag, Longint value);
extern void consAddFragmentFragment(CONSTRAINTFRAGMENT* frag,
				    const CONSTRAINTFRAGMENT* other);
extern void consAppendFragmentArrayIndex(CONSTRAINTFRAGMENT* frag, Longint value);
extern void consAppendFragmentEnumIndex(CONSTRAINTFRAGMENT* frag,
					const SYM* atom);
extern void consMultiplyFragment(CONSTRAINTFRAGMENT* frag, Longint factor);
extern void consNegateFragment(CONSTRAINTFRAGMENT* frag);
extern FRAGTYPE consGetFragmentType(const CONSTRAINTFRAGMENT* frag);
extern const SYM* consGetFragmentTypeSym(const CONSTRAINTFRAGMENT* frag);
extern Bool consIsFragmentConst(const CONSTRAINTFRAGMENT* frag);
extern Bool consIsFragmentSymbol(const CONSTRAINTFRAGMENT* frag);
extern Bool consIsFragmentBoolean(const CONSTRAINTFRAGMENT* frag);
extern const SYM* consGetFragmentSymbol(const CONSTRAINTFRAGMENT* frag);
extern Longint consGetFragmentMinVal(const CONSTRAINTFRAGMENT* frag);
extern Longint consGetFragmentMaxVal(const CONSTRAINTFRAGMENT* frag);
extern void consSetFragmentBounds(CONSTRAINTFRAGMENT* frag, Longint lower, Longint upper);
extern int consGetFragmentBitlength(const CONSTRAINTFRAGMENT* frag);
extern void consDisplayFragment(const CONSTRAINTFRAGMENT* frag);

/* CONSTRAINT */
extern CONSTRAINT* consCreate(INEQTYPE ineqtype, Longint rhs);
extern CONSTRAINT* consCreateFromFragment(const CONSTRAINTFRAGMENT* frag,
					  INEQTYPE ineqtype,
					  Longint rhs);
extern CONSTRAINT* consCreateProperty(const CONSTRAINTFRAGMENT* frag);
extern CONSTRAINT* consCreateComment(const char* comment);
extern CONSTRAINT* consCreateOpComment(const char* opname, const SYM* symres, const CONSTRAINTFRAGMENT* fragop1,
   const CONSTRAINTFRAGMENT* fragop2, const CONSTRAINTFRAGMENT* fragop3);
extern CONSTRAINT* consCreateLineComment(const Node* node);

extern void consAddCoef(CONSTRAINT* cons, COEF* coef);
extern void consDisplay(const CONSTRAINT* cons);

/* CONSTRAINTSET */
extern CONSTRAINTSET* consCreateCSet(SYMTABLE* symtable);
extern void consDisplayCSet(const CONSTRAINTSET* cset);

extern void consAddToSet(CONSTRAINTSET* cset, CONSTRAINT* cons);
extern SYM* consCreateBinarySymbol(CONSTRAINTSET* cset);
extern SYM* consCreateIntSymbol(CONSTRAINTSET* cset, Longint lower, Longint upper);
extern SYM* consCreateEnumSymbol(CONSTRAINTSET* cset);
extern SYM* consCreateBoolArraySymbol(CONSTRAINTSET* cset, int first, int last);
extern SYM* consCreateIntArraySymbol(CONSTRAINTSET* cset, int first, int last, Longint lower, Longint upper);
extern SYM* consFindSymbol(CONSTRAINTSET* cset, const char* name);
extern void consCreateEnumEqualities(CONSTRAINTSET* cset);

extern int consGetNumConstraints(const CONSTRAINTSET* cset);
extern int consGetNumVariables(const CONSTRAINTSET* cset);
extern int consAssignVariableNumbers(CONSTRAINTSET* cset);

extern void consCSetToLP(const CONSTRAINTSET* cset, int maxT, FILE* f);

#endif
