/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                           */
/*                  This file is part of the program                         */
/*        SMV2LP -- Converting SMV to Integer Linear Programs                */
/*                                                                           */
/*    Copyright (C) 2002-2005 Tobias Achterberg                              */
/*                            Thorsten Koch                                  */
/*                            Robi Malik                                     */
/*                                                                           */
/*                  2002-2005 Konrad-Zuse-Zentrum                            */
/*                            fuer Informationstechnik Berlin                */
/*                                                                           */
/*  SMV2LP is distributed under the terms of the ZIB Academic License.       */
/*                                                                           */
/*  You should have received a copy of the ZIB Academic License              */
/*  along with SMV2LP; see the file COPYING.                                 */
/*                                                                           */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/* -*-  c-basic-offset: 3 -*- */

#include <assert.h>
#include <stdio.h>
#include <math.h>

#include "def.h"

#include "message.h"
#include "memory.h"
#include "constraint.h"
#include "convert.h"
#include "node.h"


/* TYPE CHECKING ************************************************************/

/**< Check whether the given constraint fragment represents a binary value,
 *   and report an error if it does not.
 */
static void convCheckBoolean(const CONSTRAINTFRAGMENT* frag, const Node* node)
{
   char errmsg[80];

   assert(frag != NULL);
   assert(node != NULL);

   if (!consIsFragmentBoolean(frag))
   {
      sprintf(errmsg, "Boolean expression expected in line %d", node->lineno);
      errorMessage(errmsg);
      abort();
   }
}


/**< Check whether the given constraint fragment represents an integer
 *   or binary variable and report an error if it does not.
 */
static void convCheckInteger(const CONSTRAINTFRAGMENT* frag, const Node* node)
{
   char errmsg[80];

   assert(frag != NULL);
   assert(node != NULL);

   if (consGetFragmentType(frag) != FRAGTYPE_SCALAR)
   {
      sprintf(errmsg, "Numeric expression expected in line %d", node->lineno);
      errorMessage(errmsg);
      abort();
   }
}


/**< User is trying to assign value in frag2 to variable in frag1.
 *   Check whether this assignment is type-safe and issue an error if not.
 */
static void convCheckAssignmentTypes(CONSTRAINTFRAGMENT* frag1,
				     CONSTRAINTFRAGMENT* frag2,
				     const Node* node)
{
   char errmsg[80];
   const SYM* typesym1;
   const SYM* typesym2;

   assert(frag1 != NULL);
   assert(frag2 != NULL);
   assert(node != NULL);

   if (consIsFragmentConst(frag1)) {
      sprintf(errmsg, "Assigning to a constant in line %d", node->lineno);
      errorMessage(errmsg);
      abort();
   }

   switch (consGetFragmentType(frag1))
   {
   case FRAGTYPE_ENUM:
      if (consGetFragmentType(frag2) != FRAGTYPE_ENUM) {
	 sprintf(errmsg,
		 "Assigning numeric to enumeration in line %d",
		 node->lineno);
	 errorMessage(errmsg);
	 abort();
      }

      typesym1 = consGetFragmentTypeSym(frag1);
      typesym2 = consGetFragmentTypeSym(frag2);
      assert(typesym1->type == SYMTYPE_ENUMVAR);

      switch (typesym2->type)
      {
      case SYMTYPE_ATOM:
	 if (!symContainsEnumAtom(typesym1, typesym2)) {
	    sprintf(errmsg,
		    "Can't assign constant %s in line %d",
		    typesym2->name,
		    node->lineno);
	    errorMessage(errmsg);
	    abort();
	 }
	 break;
      case SYMTYPE_ENUMVAR:
	 if (!symContainsEnumEnum(typesym1, typesym2)) {
	    sprintf(errmsg,
		    "Range of right-hand side too large "
		    "in assignment in line %d",
		    node->lineno);
	    errorMessage(errmsg);
	    abort();
	 }
	 break;
      default:
	 assert(FALSE);  /* Unexpected symbol type on right-hand side! */
	 break;
      }
      break;

   case FRAGTYPE_SCALAR:
      if (consGetFragmentType(frag2) != FRAGTYPE_SCALAR) {
	 sprintf(errmsg,
		 "Assigning enumeration to numeric in line %d",
		 node->lineno);
	 errorMessage(errmsg);
	 abort();
      } else if (consGetFragmentMinVal(frag1) > consGetFragmentMinVal(frag2) ||
		 consGetFragmentMaxVal(frag1) < consGetFragmentMaxVal(frag2)) {
	 sprintf(errmsg,
		 "Range of right-hand side too large in assignment in line %d",
		 node->lineno);
	 errorMessage(errmsg);
	 consDisplayFragment(frag1);
	 consDisplayFragment(frag2);
	 abort();
      }
      break;

   case FRAGTYPE_ARRAY:
      sprintf(errmsg,
	      "Assignment of arrays not supported in line %d",
	      node->lineno);
      errorMessage(errmsg);
      abort();
      break;

   default:
      assert(FALSE);  /* Unknown fragment type! */
      break;
   }
}


/**< User is trying to compare values in frag1 and frag2 for equality.
 *   Check whether this comparison is type-safe and issue and error if not.
 */
static void convCheckEqualityTypes(CONSTRAINTFRAGMENT* frag1,
				   CONSTRAINTFRAGMENT* frag2,
				   const Node* node)
{
   char errmsg[80];

   assert(frag1 != NULL);
   assert(frag2 != NULL);
   assert(node != NULL);

   switch (consGetFragmentType(frag1))
   {
   case FRAGTYPE_ENUM:
      if (consGetFragmentType(frag2) != FRAGTYPE_ENUM) {
	 sprintf(errmsg,
		 "Comparing numeric to enumeration in line %d",
		 node->lineno);
	 errorMessage(errmsg);
	 abort();
      }
      break;
   case FRAGTYPE_SCALAR:
      if (consGetFragmentType(frag2) != FRAGTYPE_SCALAR) {
	 sprintf(errmsg,
		 "Comparing enumeration to numeric in line %d",
		 node->lineno);
	 errorMessage(errmsg);
	 abort();
      }
      break;
   case FRAGTYPE_ARRAY:
      sprintf(errmsg,
	      "Comparing of arrays not supported in line %d",
	      node->lineno);
      errorMessage(errmsg);
      abort();
      break;
   default:
      assert(FALSE);  /* Unknown fragment type! */
      break;
   }
}


/**< Find a symbol by its name.
 *   This method extracts the symbol name from the given parse tree node,
 *   and looks up the corresponding symbol in the symbol table of the
 *   given constraint set. If the symbol is not found, an error message
 *   is produced.
 */
static SYM* convFindSymbol(CONSTRAINTSET* cset, const Node* node)
{
   char* name;
   SYM* symbol;
   char errmsg[80];

   assert(cset != NULL);
   assert(node != NULL);

   name = symGetNameFromNode(node);
   symbol = consFindSymbol(cset, name);
   if (symbol == NULL) {
      sprintf(errmsg, "Undefined symbol %s in line %d", name, node->lineno);
      errorMessage(errmsg);
      abort();
   }
   freeMemoryArray(&name);

   return symbol;
}


/* CREATING FRAGMENTS *******************************************************/

/* forward declaration: */
static CONSTRAINTFRAGMENT* convCreateLinearFragment(CONSTRAINTSET* cset,
						    const Node* node,
						    TIMESTAT timestat); 


/**< Create constraints for a conjunction.
 *   This method produces constraints and returns a binary fragment,
 *   which is true if and only if <frag1> & <frag2> is true.
 *   Both input fragments must be of boolean type.
 *   The returned fragment is constant true or false, or equal to
 *   <frag1> or <frag2>, if this can be inferred from the bounds of
 *   <frag1> and <frag2>. However, it will never share any structure
 *   with <frag1> or <frag2>.
 */
static CONSTRAINTFRAGMENT* convCreateAndFragment(
   CONSTRAINTSET* cset,
   const CONSTRAINTFRAGMENT* frag1,
   const CONSTRAINTFRAGMENT* frag2,
   TIMESTAT timestat)
{
   /* SMV CODE:
    *   $$ := <exp1> & <exp2>;
    *
    * LP:
    *   $$ - <exp1> <= 0;
    *   $$ - <exp2> <= 0;
    *   $$ - <exp1> - <exp2> >= -1;
    *
    * TBD:
    * - Flatten nested AND nodes.
    */

   SYM* symbol;
   CONSTRAINTFRAGMENT* frag;
   CONSTRAINTFRAGMENT* minusfrag1;
   CONSTRAINTFRAGMENT* minusfrag2;
   CONSTRAINT* cons;

   assert(cset != NULL);
   assert(frag1 != NULL);
   assert(consGetFragmentType(frag1) == FRAGTYPE_SCALAR);
   assert(consGetFragmentMinVal(frag1) >= 0);
   assert(consGetFragmentMinVal(frag1) <= 1);
   assert(frag2 != NULL);
   assert(consGetFragmentType(frag2) == FRAGTYPE_SCALAR);
   assert(consGetFragmentMinVal(frag2) >= 0);
   assert(consGetFragmentMinVal(frag2) <= 1);

   /* Check trivial cases ... */
   if (consIsFragmentConst(frag1)) {
      if (consGetFragmentMinVal(frag1) == 0) {
	 /*  0 & <frag2>  >>  0  */
	 return consCreateFragmentConst(0);
      } else {
	 /*  1 & <frag2>  >>  <frag2>  */
	 return consCopyFragment(frag2);
      }
   } else if (consIsFragmentConst(frag2)) {
      if (consGetFragmentMinVal(frag2) == 0) {
	 /*  <frag1> & 0  >>  0  */
	 return consCreateFragmentConst(0);
      } else {
	 /*  <frag1> & 1  >>  <frag1>  */
	 return consCopyFragment(frag1);
      }
   }

   /* Create symbol ... */
   symbol = consCreateBinarySymbol(cset);

   /* add comment */
   consAddToSet(cset, consCreateOpComment("AND", symbol, frag1, frag2, NULL));

   /* Create constraints ... */
   minusfrag1 = consCopyFragment(frag1);
   consMultiplyFragment(minusfrag1, -1);
   minusfrag2 = consCopyFragment(frag2);
   consMultiplyFragment(minusfrag2, -1);

   /* $$ - <exp1> <= 0 */
   frag = consCreateFragmentVar(symbol, timestat);
   consAddFragmentFragment(frag, minusfrag1);
   cons = consCreateFromFragment(frag, INEQTYPE_LESS, 0);
   consDestroyFragment(frag);
   consAddToSet(cset, cons);

   /* $$ - <exp2> <= 0 */
   frag = consCreateFragmentVar(symbol, timestat);
   consAddFragmentFragment(frag, minusfrag2);
   cons = consCreateFromFragment(frag, INEQTYPE_LESS, 0);
   consDestroyFragment(frag);
   consAddToSet(cset, cons);

   /* $$ - <exp1> - <exp2> >= -1 */
   frag = consCreateFragmentVar(symbol, timestat);
   consAddFragmentFragment(frag, minusfrag1);
   consDestroyFragment(minusfrag1);
   consAddFragmentFragment(frag, minusfrag2);
   consDestroyFragment(minusfrag2);
   cons = consCreateFromFragment(frag, INEQTYPE_GREATER, -1);
   consDestroyFragment(frag);
   consAddToSet(cset, cons);
   
   /* Create result fragment ... */
   return consCreateFragmentVar(symbol, timestat);
}


/**< Create constraints for a disjunction.
 *   This method produces constraints and returns a binary fragment,
 *   which is true if and only if <frag1> | <frag2> is true.
 *   Both input fragments must be of boolean type.
 *   The returned fragment is constant true or false, or equal to
 *   <frag1> or <frag2>, if this can be inferred from the bounds of
 *   <frag1> and <frag2>. However, it will never share any structure
 *   with <frag1> or <frag2>.
 */
static CONSTRAINTFRAGMENT* convCreateOrFragment(
   CONSTRAINTSET* cset,
   const CONSTRAINTFRAGMENT* frag1,
   const CONSTRAINTFRAGMENT* frag2,
   TIMESTAT timestat)
{
   /* SMV CODE:
    *   $$ := <exp1> & <exp2>;
    *
    * LP:
    *   $$ - <exp1> >= 0;
    *   $$ - <exp2> >= 0;
    *   $$ - <exp1> - <exp2> <= 0;
    *
    * TBD:
    * - Flatten nested OR nodes.
    */

   SYM* symbol;
   CONSTRAINTFRAGMENT* frag;
   CONSTRAINTFRAGMENT* minusfrag1;
   CONSTRAINTFRAGMENT* minusfrag2;
   CONSTRAINT* cons;

   assert(cset != NULL);
   assert(frag1 != NULL);
   assert(consGetFragmentType(frag1) == FRAGTYPE_SCALAR);
   assert(consGetFragmentMinVal(frag1) >= 0);
   assert(consGetFragmentMinVal(frag1) <= 1);
   assert(frag2 != NULL);
   assert(consGetFragmentType(frag2) == FRAGTYPE_SCALAR);
   assert(consGetFragmentMinVal(frag2) >= 0);
   assert(consGetFragmentMinVal(frag2) <= 1);

   /* Check trivial cases ... */
   if (consIsFragmentConst(frag1)) {
      if (consGetFragmentMinVal(frag1) == 0) {
	 /*  0 | <frag2>  >>  <frag2>  */
	 return consCopyFragment(frag2);
      } else {
	 /*  1 | <frag2>  >>  1  */
	 return consCreateFragmentConst(1);
      }
   } else if (consIsFragmentConst(frag2)) {
      if (consGetFragmentMinVal(frag2) == 0) {
	 /*  <frag1> | 0  >>  <frag1>  */
	 return consCopyFragment(frag1);
      } else {
	 /*  <frag1> | 1  >>  1  */
	 return consCreateFragmentConst(1);
      }
   }

   /* Create symbol ... */
   symbol = consCreateBinarySymbol(cset);

   /* add comment */
   consAddToSet(cset, consCreateOpComment("OR", symbol, frag1, frag2, NULL));

   /* Create constraints ... */
   minusfrag1 = consCopyFragment(frag1);
   consMultiplyFragment(minusfrag1, -1);
   minusfrag2 = consCopyFragment(frag2);
   consMultiplyFragment(minusfrag2, -1);

   /* $$ - <exp1> >= 0 */
   frag = consCreateFragmentVar(symbol, timestat);
   consAddFragmentFragment(frag, minusfrag1);
   cons = consCreateFromFragment(frag, INEQTYPE_GREATER, 0);
   consDestroyFragment(frag);
   consAddToSet(cset, cons);

   /* $$ - <exp2> >= 0 */
   frag = consCreateFragmentVar(symbol, timestat);
   consAddFragmentFragment(frag, minusfrag2);
   cons = consCreateFromFragment(frag, INEQTYPE_GREATER, 0);
   consDestroyFragment(frag);
   consAddToSet(cset, cons);

   /* $$ - <exp1> - <exp2> <= 0 */
   frag = consCreateFragmentVar(symbol, timestat);
   consAddFragmentFragment(frag, minusfrag1);
   consDestroyFragment(minusfrag1);
   consAddFragmentFragment(frag, minusfrag2);
   consDestroyFragment(minusfrag2);
   cons = consCreateFromFragment(frag, INEQTYPE_LESS, 0);
   consDestroyFragment(frag);
   consAddToSet(cset, cons);
   
   /* Create result fragment ... */
   return consCreateFragmentVar(symbol, timestat);
}


/**< Create constraints for a logic equivalance.
 *   This method produces constraints and returns a binary fragment,
 *   which is true if and only if <frag1> <-> <frag2> is true.
 *   Both input fragments must be of boolean type.
 *   The returned fragment is constant true or false, or equal to
 *   <frag1> or <frag2>, if this can be inferred from the bounds of
 *   <frag1> and <frag2>. However, it will never share any structure
 *   with <frag1> or <frag2>.
 */
static CONSTRAINTFRAGMENT* convCreateIffFragment
  (CONSTRAINTSET* cset,
   const CONSTRAINTFRAGMENT* frag1,
   const CONSTRAINTFRAGMENT* frag2,
   TIMESTAT timestat)
{
   /* SMV CODE:
    *   $$ := <exp1> <-> <exp2>;
    *
    * LP:
    *   $$ - <exp1> + <exp2> <= 1
    *   $$ + <exp1> - <exp2> <= 1
    *   $$ + <exp1> + <exp2> >= 1
    *   $$ - <exp1> - <exp2> >= -1
    */

   SYM* symbol;
   CONSTRAINTFRAGMENT* frag;
   CONSTRAINTFRAGMENT* minusfrag1;
   CONSTRAINTFRAGMENT* minusfrag2;
   CONSTRAINT* cons;

   assert(cset != NULL);
   assert(frag1 != NULL);
   assert(consGetFragmentType(frag1) == FRAGTYPE_SCALAR);
   assert(consGetFragmentMinVal(frag1) >= 0);
   assert(consGetFragmentMinVal(frag1) <= 1);
   assert(frag2 != NULL);
   assert(consGetFragmentType(frag2) == FRAGTYPE_SCALAR);
   assert(consGetFragmentMinVal(frag2) >= 0);
   assert(consGetFragmentMinVal(frag2) <= 1);

   /* Check trivial cases ... */
   if (consIsFragmentConst(frag1)) {
      if (consGetFragmentMinVal(frag1) == 0) {
	 /*  0 <-> <frag2>  >>  1 - <frag2>  */
	 frag = consCopyFragment(frag2);
	 consNegateFragment(frag);
	 return frag;
      } else {
	 /*  1 <-> <frag2>  >>  <frag2>  */
	 return consCopyFragment(frag2);
      }
   } else if (consIsFragmentConst(frag2)) {
      if (consGetFragmentMinVal(frag2) == 0) {
	 /*  <frag1> <-> 0  >>  1 - <frag1>  */
	 frag = consCopyFragment(frag1);
	 consNegateFragment(frag);
	 return frag;
      } else {
	 /*  <frag1> | 1  >>  1  */
	 return consCopyFragment(frag1);
      }
   }

   /* Create symbol ... */
   symbol = consCreateBinarySymbol(cset);

   /* add comment */
   consAddToSet(cset, consCreateOpComment("IFF", symbol, frag1, frag2, NULL));

   /* Create constraints ... */
   minusfrag1 = consCopyFragment(frag1);
   consMultiplyFragment(minusfrag1, -1);
   minusfrag2 = consCopyFragment(frag2);
   consMultiplyFragment(minusfrag2, -1);

   /* Constraint: $$ - <exp1> + <exp2> <= 1 */
   frag = consCreateFragmentVar(symbol, timestat);
   consAddFragmentFragment(frag, minusfrag1);
   consAddFragmentFragment(frag, frag2);
   cons = consCreateFromFragment(frag, INEQTYPE_LESS, 1);
   consAddToSet(cset, cons);
   consDestroyFragment(frag);

   /* Constraint: $$ + <exp1> - <exp2> <= 1 */
   frag = consCreateFragmentVar(symbol, timestat);
   consAddFragmentFragment(frag, frag1);
   consAddFragmentFragment(frag, minusfrag2);
   cons = consCreateFromFragment(frag, INEQTYPE_LESS, 1);
   consAddToSet(cset, cons);
   consDestroyFragment(frag);

   /* Constraint: $$ + <exp1> + <exp2> >= 1 */
   frag = consCreateFragmentVar(symbol, timestat);
   consAddFragmentFragment(frag, frag1);
   consAddFragmentFragment(frag, frag2);
   cons = consCreateFromFragment(frag, INEQTYPE_GREATER, 1);
   consAddToSet(cset, cons);
   consDestroyFragment(frag);

   /* Constraint: $$ - <exp1> - <exp2> >= -1 */
   frag = consCreateFragmentVar(symbol, timestat);
   consAddFragmentFragment(frag, minusfrag1);
   consAddFragmentFragment(frag, minusfrag2);
   cons = consCreateFromFragment(frag, INEQTYPE_GREATER, -1);
   consAddToSet(cset, cons);
   consDestroyFragment(frag);

   /* Clean up ... */
   consDestroyFragment(minusfrag1);
   consDestroyFragment(minusfrag2);
   
   /* Create result fragment ... */
   return consCreateFragmentVar(symbol, timestat);
}

/**< creates bit expansion for fragment 'frag' and returns the newly created 'array[] of binary' symbol to store it.
 */
static SYM* convCreateBitexpansion(
   CONSTRAINTSET*      cset,
   CONSTRAINTFRAGMENT* frag,
   TIMESTAT            timestat)
{
   SYM*                bitarray;
   CONSTRAINTFRAGMENT* bitfrag;
   CONSTRAINTFRAGMENT* partfrag;
   CONSTRAINT*         cons;
   int  bitlength;
   Bool needsign;
   Longint  coef;
   int  i;

   /* get bitlength of fragment, and check, whether it can be negative */
   bitlength = consGetFragmentBitlength(frag);
   assert(bitlength > 0);
   needsign = (consGetFragmentMinVal(frag) < 0);

   bitarray = consCreateBoolArraySymbol(cset, 0, bitlength-1);

   /* create constraint fragment:  frag - 2^0b0 - 2^1b1 - ... - 2^(n-1)b[n-1]         , if frag is always non-negative
    *                              frag - 2^0b0 - ... - 2^(n-2)b[n-2] + 2^(n-1)b[n-1] , if frag can be negative
    */
   bitfrag = consCopyFragment(frag);
   coef = 1;
   for( i = 0; i < bitlength; i++ )
   {
      partfrag = consCreateFragmentVar(bitarray, timestat);
      consAppendFragmentArrayIndex(partfrag, i);
      if( needsign && i == bitlength-1 )
         consMultiplyFragment(partfrag, +coef);
      else
         consMultiplyFragment(partfrag, -coef);
      consAddFragmentFragment(bitfrag, partfrag);
      consDestroyFragment(partfrag);
      coef *= 2;
   }

   cons = consCreateFromFragment(bitfrag, INEQTYPE_EQUAL, 0);
   consAddToSet(cset, cons);

   consDestroyFragment(bitfrag);
   
   return bitarray;
}


/**< Create fragment, necessary inequalities, and necessary variables for <frag1> * <frag2>.
 */
static CONSTRAINTFRAGMENT* convCreateTimesFragment(
   CONSTRAINTSET*      cset,
   CONSTRAINTFRAGMENT* frag1,
   CONSTRAINTFRAGMENT* frag2,
   TIMESTAT            timestat)
{
   CONSTRAINT*         cons;
   CONSTRAINTFRAGMENT* frag;
   CONSTRAINTFRAGMENT* tmpfrag1;
   CONSTRAINTFRAGMENT* tmpfrag2;
   const SYM*          symfrag1 = NULL;
   const SYM*          bitarray = NULL;
   SYM*                psym     = NULL;
   Longint  frag1min;
   Longint  frag1max;
   Longint  frag2min;
   Longint  frag2max;
   Longint  pmin;
   Longint  pmax;
   int  bitlength;
   int  i;
   Longint  coef;
   Bool needsign;
   Longint  lower;
   Longint  upper;

   /* get bounds of fragments */
   frag1min = consGetFragmentMinVal(frag1);
   frag1max = consGetFragmentMaxVal(frag1);
   frag2min = consGetFragmentMinVal(frag2);
   frag2max = consGetFragmentMaxVal(frag2);

   /* is frag1 a constant? */
   if( consIsFragmentConst(frag1) )
   {
      frag = consCopyFragment(frag2);
      consMultiplyFragment(frag, frag1min);
      return frag;
   }

   /* is frag2 a constant? */
   if( consIsFragmentConst(frag2) )
   {
      frag = consCopyFragment(frag1);
      consMultiplyFragment(frag, frag2min);
      return frag;
   }

   /* both fragments are non-constant */
   
   /* add comment */
   consAddToSet(cset, consCreateOpComment("TIMES", NULL, frag1, frag2, NULL));

   /* put smaller fragment in frag1 */
   if( consGetFragmentBitlength(frag2) < consGetFragmentBitlength(frag1) )
   {
      frag = frag2;
      frag2 = frag1;
      frag1 = frag;
   }

   /* if larger fragment is already bitexpanded and smaller fragment is not, switch fragments */
   if( !(consIsFragmentSymbol(frag1) && symIsIntvarBitexpanded(consGetFragmentSymbol(frag1)))
      && consIsFragmentSymbol(frag2) && symIsIntvarBitexpanded(consGetFragmentSymbol(frag2)) )
   {
      frag = frag2;
      frag2 = frag1;
      frag1 = frag;
   }
   /* now, the fragment to be bitextended is frag1 */
   
   /* check, if bit expansion of frag1 is already existent */
   if( consIsFragmentSymbol(frag1) )
   {
      symfrag1 = consGetFragmentSymbol(frag1);
      bitarray = symGetIntvarBitarray(symfrag1);
   }

   /* construct the bit extension, if not already existent */
   if( bitarray == NULL )
      bitarray = convCreateBitexpansion(cset, frag1, timestat);

   assert(bitarray != NULL);

   /* if frag1 is symbol and not yet expanded, remember the bitarray for the symbol */
   if( symfrag1 != NULL )
   {
      assert(consIsFragmentSymbol(frag1));
      assert(symfrag1 == consGetFragmentSymbol(frag1));
      if( !symIsIntvarBitexpanded(symfrag1) )
         symSetIntvarBitarray(symfrag1, bitarray);
      assert(symGetIntvarBitarray(symfrag1) == bitarray);
   }

   /* construct additional variables p[i] to store the products b[i]*frag2 */
   assert(bitarray->value.array.first == 0);
   bitlength = bitarray->value.array.last+1;
   pmin = MIN(0, frag2min);
   pmax = MAX(0, frag2max);
   psym = consCreateIntArraySymbol(cset, 0, bitlength-1, pmin, pmax);
   assert(psym != NULL);

   /* construct inequalities necessary to store the products p[i] = b[i]*frag2 */
   for( i = 0; i < bitlength; ++i )
   {
      /* a) p[i,t] - maxP * y1b[i,t] <= 0 */
      tmpfrag1 = consCreateFragmentVar(psym, timestat);     /* tmpfrag1 = p */
      consAppendFragmentArrayIndex(tmpfrag1, i);            /* tmpfrag1 = p[i] */
      tmpfrag2 = consCreateFragmentVar(bitarray, timestat); /* tmpfrag2 = y1b */
      consAppendFragmentArrayIndex(tmpfrag2, i);            /* tmpfrag2 = y1b[i] */
      consMultiplyFragment(tmpfrag2, -pmax);                /* tmpfrag2 = - maxP * y1b[i] */
      consAddFragmentFragment(tmpfrag1, tmpfrag2);          /* tmpfrag1 = p[i] - maxP * y1b[i] */
      cons = consCreateFromFragment(tmpfrag1, INEQTYPE_LESS, 0);
      consAddToSet(cset, cons);
      consDestroyFragment(tmpfrag2);
      consDestroyFragment(tmpfrag1);

      /* b) p[i,t] - minP * y1b[i,t] >= 0 */
      tmpfrag1 = consCreateFragmentVar(psym, timestat);     /* tmpfrag1 = p */
      consAppendFragmentArrayIndex(tmpfrag1, i);            /* tmpfrag1 = p[i] */
      tmpfrag2 = consCreateFragmentVar(bitarray, timestat); /* tmpfrag2 = y1b */
      consAppendFragmentArrayIndex(tmpfrag2, i);            /* tmpfrag2 = y1b[i] */
      consMultiplyFragment(tmpfrag2, -pmin);                /* tmpfrag2 = - minP * y1b[i] */
      consAddFragmentFragment(tmpfrag1, tmpfrag2);          /* tmpfrag1 = p[i] - minP * y1b[i] */
      cons = consCreateFromFragment(tmpfrag1, INEQTYPE_GREATER, 0);
      consAddToSet(cset, cons);
      consDestroyFragment(tmpfrag2);
      consDestroyFragment(tmpfrag1);

      if( frag2max > 0 )
      {
         /* c) p[i,t] - y2[t] + (minP-maxP) * y1b[i,t] >= minP-maxP */
         tmpfrag1 = consCreateFragmentVar(psym, timestat);     /* tmpfrag1 = p */
         consAppendFragmentArrayIndex(tmpfrag1, i);            /* tmpfrag1 = p[i] */
         tmpfrag2 = consCopyFragment(frag2);                   /* tmpfrag2 = y */
         consMultiplyFragment(tmpfrag2, -1);                   /* tmpfrag2 = -y */
         consAddFragmentFragment(tmpfrag1, tmpfrag2);          /* tmpfrag1 = p[i] - y */
         consDestroyFragment(tmpfrag2);
         tmpfrag2 = consCreateFragmentVar(bitarray, timestat); /* tmpfrag2 = y1b */
         consAppendFragmentArrayIndex(tmpfrag2, i);            /* tmpfrag2 = y1b[i] */
         consMultiplyFragment(tmpfrag2, pmin-pmax);            /* tmpfrag2 = (minP-maxP) * y1b[i] */
         consAddFragmentFragment(tmpfrag1, tmpfrag2);          /* tmpfrag1 = p[i] - y + (minP-maxP) * y1b[i] */
         cons = consCreateFromFragment(tmpfrag1, INEQTYPE_GREATER, pmin-pmax);
         consAddToSet(cset, cons);
         consDestroyFragment(tmpfrag2);
         consDestroyFragment(tmpfrag1);
      }
      else
      {
         /* c') p[i,t] - y2[t] >= 0 */
         tmpfrag1 = consCreateFragmentVar(psym, timestat);     /* tmpfrag1 = p */
         consAppendFragmentArrayIndex(tmpfrag1, i);            /* tmpfrag1 = p[i] */
         tmpfrag2 = consCopyFragment(frag2);                   /* tmpfrag2 = y */
         consMultiplyFragment(tmpfrag2, -1);                   /* tmpfrag2 = -y */
         consAddFragmentFragment(tmpfrag1, tmpfrag2);          /* tmpfrag1 = p[i] - y */
         cons = consCreateFromFragment(tmpfrag1, INEQTYPE_GREATER, 0);
         consAddToSet(cset, cons);
         consDestroyFragment(tmpfrag2);
         consDestroyFragment(tmpfrag1);
      }

      if( frag2min < 0 )
      {
         /* d) p[i,t] - y2[t] + (maxP-minP) * y1b[i,t] <= maxP-minP */
         tmpfrag1 = consCreateFragmentVar(psym, timestat);     /* tmpfrag1 = p */
         consAppendFragmentArrayIndex(tmpfrag1, i);            /* tmpfrag1 = p[i] */
         tmpfrag2 = consCopyFragment(frag2);                   /* tmpfrag2 = y */
         consMultiplyFragment(tmpfrag2, -1);                   /* tmpfrag2 = -y */
         consAddFragmentFragment(tmpfrag1, tmpfrag2);          /* tmpfrag1 = p[i] - y */
         consDestroyFragment(tmpfrag2);
         tmpfrag2 = consCreateFragmentVar(bitarray, timestat); /* tmpfrag2 = y1b */
         consAppendFragmentArrayIndex(tmpfrag2, i);            /* tmpfrag2 = y1b[i] */
         consMultiplyFragment(tmpfrag2, pmax-pmin);            /* tmpfrag2 = (maxP-minP) * y1b[i] */
         consAddFragmentFragment(tmpfrag1, tmpfrag2);          /* tmpfrag1 = p[i] - y + (maxP-minP) * y1b[i] */
         cons = consCreateFromFragment(tmpfrag1, INEQTYPE_LESS, pmax-pmin);
         consAddToSet(cset, cons);
         consDestroyFragment(tmpfrag2);
         consDestroyFragment(tmpfrag1);
      }
      else
      {
         /* d') p[i,t] - y2[t] <= 0 */
         tmpfrag1 = consCreateFragmentVar(psym, timestat);     /* tmpfrag1 = p */
         consAppendFragmentArrayIndex(tmpfrag1, i);            /* tmpfrag1 = p[i] */
         tmpfrag2 = consCopyFragment(frag2);                   /* tmpfrag2 = y */
         consMultiplyFragment(tmpfrag2, -1);                   /* tmpfrag2 = -y */
         consAddFragmentFragment(tmpfrag1, tmpfrag2);          /* tmpfrag1 = p[i] - y */
         cons = consCreateFromFragment(tmpfrag1, INEQTYPE_LESS, 0);
         consAddToSet(cset, cons);
         consDestroyFragment(tmpfrag2);
         consDestroyFragment(tmpfrag1);
      }
   }

   /* construct resulting fragment:
    * a) sum_{i=0,...,bitlength-1} (2^i * p[i])                                    , if minY1 >= 0
    * b) sum_{i=0,...,bitlength-2} (2^i * p[i]) - 2^(bitlength-1) * p[bitlength-1] , if minY1 < 0
    */
   needsign = (frag1min < 0);

   frag = consCreateFragment();
   coef = 1;
   for( i = 0; i < bitlength; i++ )
   {
      tmpfrag1 = consCreateFragmentVar(psym, timestat);
      consAppendFragmentArrayIndex(tmpfrag1, i);
      if( needsign && i == bitlength-1 )
         consMultiplyFragment(tmpfrag1, -coef);
      else
         consMultiplyFragment(tmpfrag1, +coef);
      consAddFragmentFragment(frag, tmpfrag1);
      consDestroyFragment(tmpfrag1);
      coef *= 2;
   }

   /* calculate bounds of fragment */
   lower = frag1min * frag2min;
   lower = MIN(lower, frag1min * frag2max);
   lower = MIN(lower, frag1max * frag2min);
   lower = MIN(lower, frag1max * frag2max);
   upper = frag1min * frag2min;
   upper = MAX(upper, frag1min * frag2max);
   upper = MAX(upper, frag1max * frag2min);
   upper = MAX(upper, frag1max * frag2max);
   consSetFragmentBounds(frag, lower, upper);

   return frag;
}

/**< Create fragment, necessary inequalities, and necessary variables for <frag1> mod <frag2>.
 */
static CONSTRAINTFRAGMENT* convCreateModFragment(
   CONSTRAINTSET*            cset,
   const CONSTRAINTFRAGMENT* frag1,
   const CONSTRAINTFRAGMENT* frag2,
   const Node*               node,
   TIMESTAT                  timestat)
{
   /* SMV CODE:
    *   $$ := <exp1> mod <const>
    *
    * LP:
    *   <const> * $quot + $$ - <exp1> = 0;
    * where
    *   0 <= $$ <= <const> - 1;
    *   floor(min(<exp1>) / <const>) <= $quot <= floor(max(<exp1>) / <const>);
    *
    * TBD:
    * - Non-constant divisor not yet implemented.
    * - Could be implemented more efficiently with more accurate range
    *   if range of dividend is small, e.g. in x mod 32 where 42 <= x <= 48.
    */

   CONSTRAINTFRAGMENT* tmpfrag1;
   CONSTRAINTFRAGMENT* tmpfrag2;
   CONSTRAINT* cons;
   SYM* resultsymbol;
   SYM* quotsymbol;
   Longint dividend;
   Longint divisor;
   Longint minval;
   Longint maxval;
   Longint minquot;
   Longint maxquot;
   char errmsg[80];

   assert(cset != NULL);
   assert(frag1 != NULL);
   assert(frag2 != NULL);

   if( consIsFragmentConst(frag2) )
   {
      /* Divisor is constant. */

      /* Get divisor and check range */
      divisor = consGetFragmentMinVal(frag2);
      if( divisor == 0 )
      {
	 sprintf(errmsg, "Division by zero in line %d", node->lineno);
	 errorMessage(errmsg);
	 abort();
      }
      else if( divisor < 0 )
      {
	 sprintf(errmsg, "Modulo by negative constant %lld in line %d", divisor, node->lineno);
	 errorMessage(errmsg);
	 abort();
      }

      if( consIsFragmentConst(frag1) )
      {
         dividend = consGetFragmentMinVal(frag1);
         return consCreateFragmentConst(dividend % divisor);
      }

      /* Create symbols for result and quotient ... */
      resultsymbol = consCreateIntSymbol(cset, 0, divisor - 1);
      minval = consGetFragmentMinVal(frag1);
      minquot = (Longint)(floor((double)minval/(double)divisor));
      maxval = consGetFragmentMaxVal(frag1);
      maxquot = (Longint)(floor((double)maxval/(double)divisor));
      quotsymbol = consCreateIntSymbol(cset, minquot, maxquot);

      /* Create constraint: */

      /* add comment */
      consAddToSet(cset, consCreateOpComment("MOD[const]", NULL, frag1, frag2, NULL));

      /* <divisor> * <quot> + <result> - <frag1> = 0 */
      tmpfrag1 = consCopyFragment(frag1);
      consMultiplyFragment(tmpfrag1, -1);
      tmpfrag2 = consCreateFragmentVar(resultsymbol, timestat);
      consAddFragmentFragment(tmpfrag1, tmpfrag2);
      consDestroyFragment(tmpfrag2);
      tmpfrag2 = consCreateFragmentVar(quotsymbol, timestat);
      consMultiplyFragment(tmpfrag2, divisor);
      consAddFragmentFragment(tmpfrag1, tmpfrag2);
      consDestroyFragment(tmpfrag2);
      cons = consCreateFromFragment(tmpfrag1, INEQTYPE_EQUAL, 0);
      consDestroyFragment(tmpfrag1);
      consAddToSet(cset, cons);
   }
   else
   {   
      /* Divisor is not constant. */

      /* add comment */
      consAddToSet(cset, consCreateOpComment("MOD[var]", NULL, frag1, frag2, NULL));

      assert(FALSE);  /* Modulo by non-constant divisor not yet implemented! */
   }

   /* Create result fragment ... */
   return consCreateFragmentVar(resultsymbol, timestat);
}



/**< Create fragment, necessary inequalities, and necessary variables for <frag1> / <frag2>.
 *   Warning! In SMV  -1/5 == -1  instead of 0 (always round down!).
 */
static CONSTRAINTFRAGMENT* convCreateDivideFragment(
   CONSTRAINTSET*            cset,
   const CONSTRAINTFRAGMENT* frag1,
   const CONSTRAINTFRAGMENT* frag2,
   const Node*               node,
   TIMESTAT                  timestat)
{
   /* SMV CODE:
    *   $$ := <exp1> / <const>
    *
    * LP:
    *   0 <= <exp1> - <const> * $quot <= <const>-1;
    * where
    *   floor(min(<exp1>) / <const>) <= $quot <= floor(max(<exp1>) / <const>);
    *
    * TBD:
    * - Non-constant divisor not yet implemented.
    * - Could be implemented more efficiently with more accurate range
    *   if range of dividend is small, e.g. in x / 32 where 42 <= x <= 48.
    */

   CONSTRAINTFRAGMENT* tmpfrag1;
   CONSTRAINTFRAGMENT* tmpfrag2;
   SYM* quotsymbol;
   Longint dividend;
   Longint divisor;
   Longint minval;
   Longint maxval;
   Longint minquot;
   Longint maxquot;
   char errmsg[80];

   assert(cset != NULL);
   assert(frag1 != NULL);
   assert(frag2 != NULL);

   if( consIsFragmentConst(frag2) )
   {
      /* Divisor is constant. */

      /* Get divisor and check range */
      divisor = consGetFragmentMinVal(frag2);
      if( divisor == 0 )
      {
	 sprintf(errmsg, "Division by zero in line %d", node->lineno);
	 errorMessage(errmsg);
	 abort();
      }
      else if( divisor < 0 )
      {
	 sprintf(errmsg, "Division by negative constant %lld in line %d", divisor, node->lineno);
	 errorMessage(errmsg);
	 abort();
      }

      if( consIsFragmentConst(frag1) )
      {
         dividend = consGetFragmentMinVal(frag1);
         return consCreateFragmentConst((Longint)(floor((double)dividend/(double)divisor)));
      }

      /* Create symbol for quotient */
      minval = consGetFragmentMinVal(frag1);
      minquot = (Longint)(floor((double)minval/(double)divisor));
      maxval = consGetFragmentMaxVal(frag1);
      maxquot = (Longint)(floor((double)maxval/(double)divisor));
      quotsymbol = consCreateIntSymbol(cset, minquot, maxquot);

      /* Create constraint: */

      /* add comment */
      consAddToSet(cset, consCreateOpComment("DIV[const]", NULL, frag1, frag2, NULL));

      /* 0 <= <exp1> - <const> * $quot <= <const>-1 */
      tmpfrag1 = consCopyFragment(frag1);
      tmpfrag2 = consCreateFragmentVar(quotsymbol, timestat);
      consMultiplyFragment(tmpfrag2, -divisor);
      consAddFragmentFragment(tmpfrag1, tmpfrag2);
      consDestroyFragment(tmpfrag2);
      consAddToSet(cset, consCreateFromFragment(tmpfrag1, INEQTYPE_GREATER, 0));
      consAddToSet(cset, consCreateFromFragment(tmpfrag1, INEQTYPE_LESS, divisor-1));
      consDestroyFragment(tmpfrag1);
   }
   else
   {   
      /* Divisor is not constant. */

      /* add comment */
      consAddToSet(cset, consCreateOpComment("DIV[var]", NULL, frag1, frag2, NULL));

      assert(FALSE);  /* Modulo by non-constant divisor not yet implemented! */
   }

   /* Create result fragment ... */
   return consCreateFragmentVar(quotsymbol, timestat);
}



/**< Helper function to create constraints for an inequality.
 *   This method should not be called directly!
 *   Use convCreateLessFragment(), convCreateLessequalFragment(), ... instead!
 *   This method produces constraints and returns a binary fragment,
 *   which is true if and only if <frag1> + <strict> <= <frag2>.
 *   Both input fragments must be of scalar type.
 *   The returned fragment is constant true or false, if this
 *   can be inferred from the bounds of <frag1> and <frag2>,
 *   otherwise a new variable is created.
 */
static CONSTRAINTFRAGMENT* convCreateComparisonFragment
  (CONSTRAINTSET* cset,
   const CONSTRAINTFRAGMENT* frag1,
   const CONSTRAINTFRAGMENT* frag2,
   int strict,
   TIMESTAT timestat)
{
   /* SMV CODE:
    *   $$ := <exp1> + strict <= <exp2>
    *
    * LP:
    *   <exp2> - <exp1> - M1 * $$ <= strict - 1
    *   <exp1> - <exp2> + M2 * $$ <= M2 - strict
    * where
    *   M1 = max(<exp2> - <exp1> - strict + 1)
    *   M2 = max(<exp1> - <exp2> + strict)
    */

   CONSTRAINT* cons;
   CONSTRAINTFRAGMENT* frag;
   CONSTRAINTFRAGMENT* partfrag;
   SYM* symbol;
   Longint minval1;
   Longint maxval1;
   Longint minval2;
   Longint maxval2;
   Longint maxdiff;

   assert(cset != NULL);
   assert(frag1 != NULL);
   assert(consGetFragmentType(frag1) == FRAGTYPE_SCALAR);
   assert(frag2 != NULL);
   assert(consGetFragmentType(frag2) == FRAGTYPE_SCALAR);
   assert(strict == 0 || strict == 1);

   /* Compute bounds and check trivial cases ... */
   minval1 = consGetFragmentMinVal(frag1);
   maxval1 = consGetFragmentMaxVal(frag1);
   minval2 = consGetFragmentMinVal(frag2);
   maxval2 = consGetFragmentMaxVal(frag2);
   if (maxval1 + strict <= minval2) {
      return consCreateFragmentConst(1);
   } else if (minval1 + strict > maxval2) {
      return consCreateFragmentConst(0);
   }

   /* Check for boolean case ... */
   if (consIsFragmentBoolean(frag1) && consIsFragmentBoolean(frag2)) {
      /* <exp1> <= <exp2>  >>  !<exp1> | <exp2> */
      /* <exp1> < <exp2>   >>  !<exp1> & <exp2> */
      partfrag = consCopyFragment(frag1);
      consNegateFragment(partfrag);
      if (strict == 0) {
	 frag = convCreateOrFragment(cset, partfrag, frag2, timestat);
      } else {
	 frag = convCreateAndFragment(cset, partfrag, frag2, timestat);
      }
      consDestroyFragment(partfrag);
      return frag;
   }

   /* Create new variable ... */
   symbol = consCreateBinarySymbol(cset);

   /* add comment */
   if( strict == 0 )
      consAddToSet(cset, consCreateOpComment("LESSEQUAL", symbol, frag1, frag2, NULL));
   else
      consAddToSet(cset, consCreateOpComment("LESS", symbol, frag1, frag2, NULL));

   /* Create constraints ... */
   /* <exp2> - <exp1> - M1 * $$ <= strict - 1,
    * where M1 = max(<exp2> - <exp1> - strict + 1) */
   maxdiff = maxval2 - minval1 - strict + 1;
   frag = consCopyFragment(frag2);
   partfrag = consCopyFragment(frag1);
   consMultiplyFragment(partfrag, -1);
   consAddFragmentFragment(frag, partfrag);
   consDestroyFragment(partfrag);
   partfrag = consCreateFragmentVar(symbol, timestat);
   consMultiplyFragment(partfrag, -maxdiff);
   consAddFragmentFragment(frag, partfrag);
   consDestroyFragment(partfrag);
   cons = consCreateFromFragment(frag, INEQTYPE_LESS, strict - 1);
   consDestroyFragment(frag);
   consAddToSet(cset, cons);

   /* <exp1> - <exp2> + M2 * $$ <= M2 - strict
    * where M2 = max(<exp1> - <exp2> + strict) */
   maxdiff = maxval1 - minval2 + strict;
   frag = consCopyFragment(frag1);
   partfrag = consCopyFragment(frag2);
   consMultiplyFragment(partfrag, -1);
   consAddFragmentFragment(frag, partfrag);
   consDestroyFragment(partfrag);
   partfrag = consCreateFragmentVar(symbol, timestat);
   consMultiplyFragment(partfrag, maxdiff);
   consAddFragmentFragment(frag, partfrag);
   consDestroyFragment(partfrag);
   cons = consCreateFromFragment(frag, INEQTYPE_LESS, maxdiff - strict);
   consDestroyFragment(frag);
   consAddToSet(cset, cons);

   /* Create result fragment ... */
   return consCreateFragmentVar(symbol, timestat);
}


/**< Create constraints for a  <left> < <right>  inequality.
 */
static CONSTRAINTFRAGMENT* convCreateLessFragment(
   CONSTRAINTSET* cset,
   const CONSTRAINTFRAGMENT* frag1,
   const CONSTRAINTFRAGMENT* frag2,
   TIMESTAT timestat)
{
   return convCreateComparisonFragment(cset, frag1, frag2, 1, timestat);
}


/**< Create constraints for a  <left> <= <right>  inequality.
 */
static CONSTRAINTFRAGMENT* convCreateLessequalFragment(
   CONSTRAINTSET* cset,
   const CONSTRAINTFRAGMENT* frag1,
   const CONSTRAINTFRAGMENT* frag2,
   TIMESTAT timestat)
{
   return convCreateComparisonFragment(cset, frag1, frag2, 0, timestat);
}


/**< Create constraints for a  <left> > <right>  inequality.
 */
static CONSTRAINTFRAGMENT* convCreateGreaterFragment(
   CONSTRAINTSET* cset,
   const CONSTRAINTFRAGMENT* frag1,
   const CONSTRAINTFRAGMENT* frag2,
   TIMESTAT timestat)
{
   return convCreateComparisonFragment(cset, frag2, frag1, 1, timestat);
}


/**< Create constraints for a  <left> >= <right>  inequality.
 */
static CONSTRAINTFRAGMENT* convCreateGreaterequalFragment(
   CONSTRAINTSET* cset,
   const CONSTRAINTFRAGMENT* frag1,
   const CONSTRAINTFRAGMENT* frag2,
   TIMESTAT timestat)
{
   return convCreateComparisonFragment(cset, frag2, frag1, 0, timestat);
}


/**< Create constraints for an equality.
 *   This method produces constraints and returns a binary fragment,
 *   which is true if and only if <left> = <right>.
 *   <left> and <right> must be of the same type, otherwise an
 *   appropriate error message is issued.
 *   The returned fragment is constant true or false, if this
 *   can be inferred from the bounds resp. values of <left> and <right>,
 *   otherwise a new variable is created.
 */
static CONSTRAINTFRAGMENT* convCreateEqualityFragment(
   CONSTRAINTSET* cset,
   const Node* left,
   const Node* right,
   TIMESTAT timestat)
{
   /* SMV CODE:
    *   $$ := <exp1> = <exp2>
    *
    * LP (for binary expression type)
    *   $$ := <exp1> <-> <exp2>
    *
    * LP (for integer expression type)
    *   $$ := <exp1> <= <exp2> & <exp1> >= <exp2>
    *
    * LP (for enumeration types if <exp2> is constant):
    *   $$ := <exp1>[<exp2>]
    *
    * LP (for enumeration types if both expressions are non-constant):
    *   $$ + <exp1>[atom] - <exp2>[atom] <= 1,
    *   $$ - <exp1>[atom] + <exp2>[atom] <= 1,
    *   $$ - <exp2>[atom] - <exp1>[atom] >= -1,
    *                    for each atom in <exp1> \cup <exp2>
    *
    * TBD:
    * - Comparison of enumerations could do slightly better if intersection
    *   of types is found to be empty.
    * - Comparison of arrays?
    */

   CONSTRAINTFRAGMENT* frag = NULL;
   CONSTRAINTFRAGMENT* frag1 = NULL;
   CONSTRAINTFRAGMENT* frag2 = NULL;
   CONSTRAINTFRAGMENT* fragle = NULL;
   CONSTRAINTFRAGMENT* fragge = NULL;
   CONSTRAINTFRAGMENT* atomfrag1 = NULL;
   CONSTRAINTFRAGMENT* atomfrag2 = NULL;
   CONSTRAINTFRAGMENT* minusatomfrag1 = NULL;
   CONSTRAINTFRAGMENT* minusatomfrag2 = NULL;
   CONSTRAINT* cons = NULL;
   SYM* resultsym = NULL;
   const SYM* typesym1 = NULL;
   const SYM* typesym2 = NULL;
   SYM* typeunion = NULL;
   SYM* atom = NULL;
   SYMTABLE* symtable = NULL;

   assert(cset != NULL);
   assert(left != NULL);
   assert(right != NULL);

   /* add comment */
   consAddToSet(cset, consCreateComment("  EQUALITY"));

   /* Get subterms and check types ... */
   frag1 = convCreateLinearFragment(cset, left, timestat);
   frag2 = convCreateLinearFragment(cset, right, timestat);
   convCheckEqualityTypes(frag1, frag2, left);

   switch (consGetFragmentType(frag1))
   {
   case FRAGTYPE_SCALAR:
      if (consIsFragmentBoolean(frag1) && consIsFragmentBoolean(frag2)) {
	 /* Return: <exp1> <-> <exp2> */
	 return convCreateIffFragment(cset, frag1, frag2, timestat);
      } else {
	 /* Return: <exp1> <= <exp2> & <exp1> >= <exp2> */
	 fragle = convCreateLessequalFragment(cset, frag1, frag2, timestat);
	 fragge = convCreateGreaterequalFragment(cset, frag1, frag2, timestat);
	 frag = convCreateAndFragment(cset, fragle, fragge, timestat);
	 consDestroyFragment(fragle);
	 consDestroyFragment(fragge);
      }
      break;
   case FRAGTYPE_ENUM:
      typesym1 = consGetFragmentTypeSym(frag1);
      typesym2 = consGetFragmentTypeSym(frag2);
      if (consIsFragmentConst(frag1) && consIsFragmentConst(frag2)) {
	 if (typesym1 == typesym2) {
	    frag = consCreateFragmentConst(1);
	 } else {
	    frag = consCreateFragmentConst(0);
	 }
      } else if (!consIsFragmentConst(frag1) && consIsFragmentConst(frag2)) {
	 /* Return: <exp1>[<exp2>] */
	 frag = frag1;
	 frag1 = NULL;
	 consAppendFragmentEnumIndex(frag, typesym2);
      } else if (consIsFragmentConst(frag1) && !consIsFragmentConst(frag2)) {
	 /* Return: <exp2>[<exp1>] */
	 frag = frag2;
	 frag2 = NULL;
	 consAppendFragmentEnumIndex(frag, typesym1);
      } else {
	 resultsym = consCreateBinarySymbol(cset);
	 typeunion = symCreateDeepCopy(typesym1);
	 symUniteEnum(typeunion, typesym2);
	 for (symtable = typeunion->value.enumvar.elemlist;
	      symtable != NULL;
	      symtable = symtable->next) {
	    atom = symtable->sym;

	    /* Prepare creation of constraints for each atom ... */
	    if (symContainsEnumAtom(typesym1, atom)) {
	       atomfrag1 = consCopyFragment(frag1);
	       consAppendFragmentEnumIndex(atomfrag1, atom);
	    } else {
	       atomfrag1 = consCreateFragmentConst(0);
	    }
	    if (symContainsEnumAtom(typesym2, atom)) {
	       atomfrag2 = consCopyFragment(frag2);
	       consAppendFragmentEnumIndex(atomfrag2, atom);
	    } else {
	       atomfrag2 = consCreateFragmentConst(0);
	    }
	    minusatomfrag1 = consCopyFragment(atomfrag1);
	    consMultiplyFragment(minusatomfrag1, -1);
	    minusatomfrag2 = consCopyFragment(atomfrag2);
	    consMultiplyFragment(minusatomfrag2, -1);

	    /* Constraint: $$ + <exp1>[atom] - <exp2>[atom] <= 1 */
	    frag = consCreateFragmentVar(resultsym, timestat);
	    consAddFragmentFragment(frag, atomfrag1);
	    consAddFragmentFragment(frag, minusatomfrag2);
	    cons = consCreateFromFragment(frag, INEQTYPE_LESS, 1);
	    consDestroyFragment(frag);
	    consAddToSet(cset, cons);

	    /* Constraint: $$ - <exp1>[atom] + <exp2>[atom] <= 1 */
	    frag = consCreateFragmentVar(resultsym, timestat);
	    consAddFragmentFragment(frag, minusatomfrag1);
	    consAddFragmentFragment(frag, atomfrag2);
	    cons = consCreateFromFragment(frag, INEQTYPE_LESS, 1);
	    consDestroyFragment(frag);
	    consAddToSet(cset, cons);

	    /* Constraint: $$ - <exp1>[atom] - <exp2>[atom] >= -1 */
	    frag = consCreateFragmentVar(resultsym, timestat);
	    consAddFragmentFragment(frag, minusatomfrag1);
	    consAddFragmentFragment(frag, minusatomfrag2);
	    cons = consCreateFromFragment(frag, INEQTYPE_GREATER, -1);
	    consDestroyFragment(frag);
	    consAddToSet(cset, cons);

	    /* Clean up ... */
	    consDestroyFragment(atomfrag1);
	    consDestroyFragment(atomfrag2);
	    consDestroyFragment(minusatomfrag1);
	    consDestroyFragment(minusatomfrag2);

	    /* Return: $$ */
	    frag = consCreateFragmentVar(resultsym, timestat);
	 }
	 symDestroy(typeunion);
      }
      break;
   default:
      assert(FALSE);  /* Unknown type of comparison! */
      break;
   }

   /* Clean up ... */
   consDestroyFragment(frag1);
   consDestroyFragment(frag2);
   
   return frag;   
}


static CONSTRAINTFRAGMENT* convCreateCaseFragment(CONSTRAINTSET* cset,
						  const Node* node,
						  TIMESTAT timestat)
{
   /* SMV CODE:
    *   $$ := case
    *           <cond> : <then>;
    *           1      : <else>;
    *         esac;
    *
    * PARSE TREE STRUCTURE:
    *   NODE_CASE
    *   +- NODE_COLON
    *   |  +- <code>  (must have boolean result type)
    *   |  +- <then>
    *   +- <else>  (must be NODE_CASE or NODE_TRUEEXP)
    *
    * LP (for integer or binary result type):
    *   <then> - $$ + maxdiff(then,$) * <cond> <= maxdiff(then,$);
    *   $$ - <then> + maxdiff($,then) * <cond> <= maxdiff($,then);
    *   <else> - $$ - maxdiff(else,$) * <cond> <= 0;
    *   $$ - <else> - maxdiff($,else) * <cond> <= 0;
    * where
    *   maxdiff(exp,$) = max(<exp>) - min(<then>,<else>);
    *   maxdiff($,exp) = max(<then>,<else>) - min(<exp>);
    *
    * LP (for enumeration result type):
    *   $$[atom] - <then>[atom] - <cond> >= -1,
    *   $$[atom] - <else>[atom] + <cond> >= 0,
    *              for each atom in <then> \cup <else>
    *
    * TBD:
    * - Collect nested cases.
    * - Handle conjunctive conditions.
    * - Recognize exclusive conditions.
    * - Check that each one condition in nested cases is satisfied for
    *   enumeration result type.
    * - Some unused constraints should be deleted for certain trivial cases.
    * - More accurate computation of minimum and maximum for scalar
    *   result type using conditions.
    * - Array result type?
    */

   Node* left = NULL;
   Node* right = NULL;
   CONSTRAINTFRAGMENT* condfrag = NULL;
   CONSTRAINTFRAGMENT* thenfrag = NULL;
   CONSTRAINTFRAGMENT* elsefrag = NULL;
   CONSTRAINTFRAGMENT* partfrag = NULL;
   CONSTRAINTFRAGMENT* frag = NULL;
   CONSTRAINT* cons = NULL;
   SYM* symbol = NULL;
   const SYM* thentype = NULL;
   const SYM* elsetype = NULL;
   const SYM* atom = NULL;
   SYMTABLE* list = NULL;
   Longint thenmin;
   Longint thenmax;
   Longint elsemin;
   Longint elsemax; 
   Longint totalmin;
   Longint totalmax;
   char errmsg[1024];

   assert(cset != NULL);
   assert(node != NULL);
   assert(node->type == NODE_CASE);

   consAddToSet(cset, consCreateLineComment(node));

   left = node->left;
   assert(left != NULL);
   assert(left->type == NODE_COLON);
   assert(left->left != NULL);
   assert(left->right != NULL);

   right = node->right;
   assert(right != NULL);
   assert(right->type == NODE_CASE || right->type == NODE_TRUEEXP);

   /* Get condition ... */
   condfrag = convCreateLinearFragment(cset, left->left, timestat);
   convCheckBoolean(condfrag, left->left);
   
   /* If condition is constant, simply return then resp. else branch ... */
   if (consIsFragmentConst(condfrag)) {
      if (consGetFragmentMinVal(condfrag) == 1) {
	 frag = convCreateLinearFragment(cset, left->right, timestat);
      } else {
	 frag = convCreateLinearFragment(cset, right, timestat);
      }
      consDestroyFragment(condfrag);
      return frag;
   }

   /* Get branches ... */
   thenfrag = convCreateLinearFragment(cset, left->right, timestat);
   elsefrag = convCreateLinearFragment(cset, right, timestat);

   /* add comment */
   consAddToSet(cset, consCreateOpComment("CASE", NULL, condfrag, thenfrag, elsefrag));

   /* Split on result type ... */
   switch (consGetFragmentType(thenfrag))
   {
   case FRAGTYPE_SCALAR:
      /* Check types ... */
      if (consGetFragmentType(elsefrag) != FRAGTYPE_SCALAR) {
	 sprintf(errmsg, "Type mismatch in cases in line %d", node->lineno);
	 errorMessage(errmsg);
	 abort();
      }

      /* Compute minimum and maximum, then create symbol for result ... */
      thenmin = consGetFragmentMinVal(thenfrag);
      thenmax = consGetFragmentMaxVal(thenfrag);
      elsemin = consGetFragmentMinVal(elsefrag);
      elsemax = consGetFragmentMaxVal(elsefrag);
      totalmin = MIN(thenmin, elsemin);
      totalmax = MAX(thenmax, elsemax);
      symbol = consCreateIntSymbol(cset, totalmin, totalmax);
      
      /* Create constraints ... */
      /* <then> - $$ + maxdiff(then,$) * <cond> <= maxdiff(then,$) */
      frag = consCopyFragment(thenfrag);
      partfrag = consCreateFragmentVar(symbol, timestat);
      consMultiplyFragment(partfrag, -1);
      consAddFragmentFragment(frag, partfrag);
      consDestroyFragment(partfrag);
      partfrag = consCopyFragment(condfrag);
      consMultiplyFragment(partfrag, thenmax - totalmin);
      consAddFragmentFragment(frag, partfrag);
      consDestroyFragment(partfrag);
      cons = consCreateFromFragment(frag, INEQTYPE_LESS, thenmax - totalmin);
      consDestroyFragment(frag);
      consAddToSet(cset, cons);
      
      /* $$ - <then> + maxdiff($,then) * <cond> <= maxdiff($,then) */
      frag = consCreateFragmentVar(symbol, timestat);
      partfrag = consCopyFragment(thenfrag);
      consMultiplyFragment(partfrag, -1);
      consAddFragmentFragment(frag, partfrag);
      consDestroyFragment(partfrag);
      partfrag = consCopyFragment(condfrag);
      consMultiplyFragment(partfrag, totalmax - thenmin);
      consAddFragmentFragment(frag, partfrag);
      consDestroyFragment(partfrag);
      cons = consCreateFromFragment(frag, INEQTYPE_LESS, totalmax - thenmin);
      consDestroyFragment(frag);
      consAddToSet(cset, cons);
      
      /* <else> - $$ - maxdiff(else,$) * <cond> <= 0 */
      frag = consCopyFragment(elsefrag);
      partfrag = consCreateFragmentVar(symbol, timestat);
      consMultiplyFragment(partfrag, -1);
      consAddFragmentFragment(frag, partfrag);
      consDestroyFragment(partfrag);
      partfrag = consCopyFragment(condfrag);
      consMultiplyFragment(partfrag, -elsemax + totalmin);
      consAddFragmentFragment(frag, partfrag);
      consDestroyFragment(partfrag);
      cons = consCreateFromFragment(frag, INEQTYPE_LESS, 0);
      consDestroyFragment(frag);
      consAddToSet(cset, cons);
      
      /* $$ - <else> - maxdiff($,else) * <cond> <= 0 */
      frag = consCreateFragmentVar(symbol, timestat);
      partfrag = consCopyFragment(elsefrag);
      consMultiplyFragment(partfrag, -1);
      consAddFragmentFragment(frag, partfrag);
      consDestroyFragment(partfrag);
      partfrag = consCopyFragment(condfrag);
      consMultiplyFragment(partfrag, -totalmax + elsemin);
      consAddFragmentFragment(frag, partfrag);
      consDestroyFragment(partfrag);
      cons = consCreateFromFragment(frag, INEQTYPE_LESS, 0);
      consDestroyFragment(frag);
      consAddToSet(cset, cons);
      break;

   case FRAGTYPE_ENUM:
      /* TBD - If else branch is constant true, now assuming condition
       * to be true and return then branch. */
      if (right->type == NODE_TRUEEXP) {
	 consDestroyFragment(condfrag);
	 consDestroyFragment(elsefrag);
	 return thenfrag;
      }

      /* Get types from branches, then create symbol for result ... */
      thentype = consGetFragmentTypeSym(thenfrag);
      elsetype = consGetFragmentTypeSym(elsefrag);
      symbol = consCreateEnumSymbol(cset);
      switch (thentype->type)
      {
      case SYMTYPE_ATOM:
	 symAddEnumEntry(symbol, (SYM*) thentype);
	 break;
      case SYMTYPE_ENUMVAR:
	 symUniteEnum(symbol, thentype);
	 break;
      default:
	 assert(FALSE);  /* Unexpected symbol type! */
      }
      switch (elsetype->type)
      {
      case SYMTYPE_ATOM:
	 symAddEnumEntry(symbol, (SYM*) elsetype);
	 break;
      case SYMTYPE_ENUMVAR:
	 symUniteEnum(symbol, elsetype);
	 break;
      default:
	 assert(FALSE);  /* Unexpected symbol type! */
      }
      assert(symbol->value.enumvar.numelem > 0);

      /* If the result type has only one element, just return a constant ...
       * TBD: Some cleaning up might be appropriate first ... */
      if (symbol->value.enumvar.numelem == 1) {
	 symbol = symbol->value.enumvar.elemlist->sym;
	 break;
      }

      /* Create constraints for each atom ... */
      for (list = symbol->value.enumvar.elemlist;
	   list != NULL;
	   list = list->next) {
	 atom = list->sym;
	 frag = consCreateFragmentVar(symbol, timestat);
	 consAppendFragmentEnumIndex(frag, atom);
	 
	 /* Constraint: $$[atom] - <then>[atom] - <cond> >= -1 */
	 if (thentype->type == SYMTYPE_ATOM) {
	    partfrag = consCreateFragmentConst(thentype == atom ? 1 : 0);
	 } else if (!symContainsEnumAtom(thentype, atom)) {
	    partfrag = consCreateFragmentConst(0);
	 } else {
	    partfrag = consCopyFragment(thenfrag);
	    consAppendFragmentEnumIndex(partfrag, atom);
	 }
	 consMultiplyFragment(partfrag, -1);
	 consAddFragmentFragment(frag, partfrag);
	 consDestroyFragment(partfrag);
	 partfrag = consCopyFragment(condfrag);
	 consMultiplyFragment(partfrag, -1);
	 consAddFragmentFragment(frag, partfrag);
	 consDestroyFragment(partfrag);
	 cons = consCreateFromFragment(frag, INEQTYPE_GREATER, 0);
	 consDestroyFragment(frag);
	 consAddToSet(cset, cons);

	 /* Constraint: $$[atom] - <else>[atom] + <cond> >= 0 */
	 if (elsetype->type == SYMTYPE_ATOM) {
	    partfrag = consCreateFragmentConst(elsetype == atom ? 1 : 0);
	 } else if (!symContainsEnumAtom(elsetype, atom)) {
	    partfrag = consCreateFragmentConst(0);
	 } else {
	    partfrag = consCopyFragment(elsefrag);
	    consAppendFragmentEnumIndex(partfrag, atom);
	 }
	 consMultiplyFragment(partfrag, -1);
	 consAddFragmentFragment(frag, partfrag);
	 consDestroyFragment(partfrag);
	 consAddFragmentFragment(frag, condfrag);
	 cons = consCreateFromFragment(frag, INEQTYPE_GREATER, 0);
	 consDestroyFragment(frag);
	 consAddToSet(cset, cons);
      }
      break;

   default:
      assert(FALSE);  /* Cases not yet implemented for result type! */
      break;
   }

   /* Clean up ... */
   consDestroyFragment(condfrag);
   consDestroyFragment(thenfrag);
   consDestroyFragment(elsefrag);

   /* Create result fragment ... */
   return consCreateFragmentVar(symbol, timestat);
}


static CONSTRAINTFRAGMENT* convCreateLinearFragment(CONSTRAINTSET* cset,
						    const Node* node,
						    TIMESTAT timestat)
{
   SYM* symbol;
   const SYM* typesym;
   CONSTRAINTFRAGMENT* frag;
   CONSTRAINTFRAGMENT* frag1;
   CONSTRAINTFRAGMENT* frag2;
   char errmsg[1024];

   assert(cset != NULL);
   assert(node != NULL);

   consAddToSet(cset, consCreateLineComment(node));
   
   switch(node->type)
   {
   case NODE_FALSEEXP:
      return consCreateFragmentConst(0);
   case NODE_TRUEEXP: 
      return consCreateFragmentConst(1);
   case NODE_NUMBER:
      return consCreateFragmentConst(node->value.numb);
   case NODE_DOT:
   case NODE_ATOM:
      symbol = convFindSymbol(cset, node);
      if (symbol->type == SYMTYPE_DEFINE) {
	 return convCreateLinearFragment
	    (cset, symbol->value.define.subnode, timestat);
      } else {
	 return consCreateFragmentVar(symbol, timestat);
      }
   case NODE_ARRAY:
      frag = convCreateLinearFragment(cset, node->left, timestat);
      frag2 = convCreateLinearFragment(cset, node->right, timestat);
      if (consGetFragmentType(frag) != FRAGTYPE_ARRAY) {
	 sprintf(errmsg,
		 "Left of [] is not an array in line %d",
		 node->lineno);
	 errorMessage(errmsg);
	 abort();
      } else if (consGetFragmentType(frag2) != FRAGTYPE_SCALAR) {
	 sprintf(errmsg,
		 "Array index is not a number in line %d",
		 node->lineno);
	 errorMessage(errmsg);
	 abort();
      } else {
	 typesym = consGetFragmentTypeSym(frag);
	 assert(typesym->type == SYMTYPE_ARRAY);
	 if (consGetFragmentMinVal(frag2) < typesym->value.array.first ||
	     consGetFragmentMaxVal(frag2) > typesym->value.array.last) {
	    sprintf(errmsg,
		    "Array index out of bounds in line %d",
		    node->lineno);
	    errorMessage(errmsg);
	    abort();
	 }
      }
      if (consIsFragmentConst(frag2)) {
	 consAppendFragmentArrayIndex(frag, consGetFragmentMinVal(frag2));
	 consDestroyFragment(frag2);
	 return frag;
      } else {
	 assert(FALSE);  /* Non-constant array index not yet implemented! */
      }
      break;
   case NODE_UMINUS:
      frag = convCreateLinearFragment(cset, node->left, timestat);
      convCheckInteger(frag, node->left);
      consMultiplyFragment(frag, -1);
      return frag;
   case NODE_NEXT: 
      if (timestat == TIMESTAT_ACTUAL) {
	 return convCreateLinearFragment(cset, node->left, TIMESTAT_NEXT);
      } else {
	 sprintf(errmsg, "Nested next() encountered in line %d", node->lineno);
	 errorMessage(errmsg);
	 abort();
      }
      break;
   case NODE_NOT:
      frag = convCreateLinearFragment(cset, node->left, timestat);
      convCheckBoolean(frag, node->left);
      consNegateFragment(frag);
      return frag;
   case NODE_AND:
      frag1 = convCreateLinearFragment(cset, node->left, timestat);
      convCheckBoolean(frag1, node->left);
      frag2 = convCreateLinearFragment(cset, node->right, timestat);
      convCheckBoolean(frag2, node->right);
      frag = convCreateAndFragment(cset, frag1, frag2, timestat);
      consDestroyFragment(frag1);
      consDestroyFragment(frag2);
      return frag;
   case NODE_OR:
      frag1 = convCreateLinearFragment(cset, node->left, timestat);
      convCheckBoolean(frag1, node->left);
      frag2 = convCreateLinearFragment(cset, node->right, timestat);
      convCheckBoolean(frag2, node->right);
      frag = convCreateOrFragment(cset, frag1, frag2, timestat);
      consDestroyFragment(frag1);
      consDestroyFragment(frag2);
      return frag;
   case NODE_IMPLIES:
      frag1 = convCreateLinearFragment(cset, node->left, timestat);
      convCheckBoolean(frag1, node->left);
      consNegateFragment(frag1);
      frag2 = convCreateLinearFragment(cset, node->right, timestat);
      convCheckBoolean(frag2, node->right);
      frag = convCreateOrFragment(cset, frag1, frag2, timestat);
      consDestroyFragment(frag1);
      consDestroyFragment(frag2);
      return frag;
   case NODE_IFF:
   case NODE_XNOR:
      frag1 = convCreateLinearFragment(cset, node->left, timestat);
      convCheckBoolean(frag1, node->left);
      frag2 = convCreateLinearFragment(cset, node->right, timestat);
      convCheckBoolean(frag2, node->right);
      frag = convCreateIffFragment(cset, frag1, frag2, timestat);
      consDestroyFragment(frag1);
      consDestroyFragment(frag2);
      return frag;
   case NODE_XOR:
      frag1 = convCreateLinearFragment(cset, node->left, timestat);
      convCheckBoolean(frag1, node->left);
      frag2 = convCreateLinearFragment(cset, node->right, timestat);
      convCheckBoolean(frag2, node->right);
      consNegateFragment(frag2);
      frag = convCreateIffFragment(cset, frag1, frag2, timestat);
      consDestroyFragment(frag1);
      consDestroyFragment(frag2);
      return frag;
   case NODE_PLUS:
      frag = convCreateLinearFragment(cset, node->left, timestat);
      convCheckInteger(frag, node->left);
      frag2 = convCreateLinearFragment(cset, node->right, timestat);
      convCheckInteger(frag2, node->right);
      consAddFragmentFragment(frag, frag2);
      consDestroyFragment(frag2);
      return frag;
   case NODE_MINUS:
      frag = convCreateLinearFragment(cset, node->left, timestat);
      convCheckInteger(frag, node->left);
      frag2 = convCreateLinearFragment(cset, node->right, timestat);
      convCheckInteger(frag2, node->right);
      consMultiplyFragment(frag2, -1);
      consAddFragmentFragment(frag, frag2);
      consDestroyFragment(frag2);
      return frag;
   case NODE_TIMES:
      frag1 = convCreateLinearFragment(cset, node->left, timestat);
      convCheckInteger(frag1, node->left);
      frag2 = convCreateLinearFragment(cset, node->right, timestat);
      convCheckInteger(frag2, node->right);
      frag = convCreateTimesFragment(cset, frag1, frag2, timestat);
      consDestroyFragment(frag2);
      consDestroyFragment(frag1);
      return frag;
   case NODE_DIVIDE:
      frag1 = convCreateLinearFragment(cset, node->left, timestat);
      convCheckInteger(frag1, node->left);
      frag2 = convCreateLinearFragment(cset, node->right, timestat);
      convCheckInteger(frag2, node->right);
      frag = convCreateDivideFragment(cset, frag1, frag2, node, timestat);
      consDestroyFragment(frag2);
      consDestroyFragment(frag1);
      return frag;
   case NODE_MOD:
      frag1 = convCreateLinearFragment(cset, node->left, timestat);
      convCheckInteger(frag1, node->left);
      frag2 = convCreateLinearFragment(cset, node->right, timestat);
      convCheckInteger(frag2, node->right);
      frag = convCreateModFragment(cset, frag1, frag2, node, timestat);
      consDestroyFragment(frag2);
      consDestroyFragment(frag1);
      return frag;
   case NODE_EQUAL:
      return convCreateEqualityFragment(cset, node->left, node->right, timestat);
   case NODE_NOTEQUAL:
      frag = convCreateEqualityFragment(cset, node->left, node->right, timestat);
      consNegateFragment(frag);
      return frag;
   case NODE_LT:
      frag1 = convCreateLinearFragment(cset, node->left, timestat);
      convCheckInteger(frag1, node->left);
      frag2 = convCreateLinearFragment(cset, node->right, timestat);
      convCheckInteger(frag2, node->right);
      frag = convCreateLessFragment(cset, frag1, frag2, timestat);
      consDestroyFragment(frag1);
      consDestroyFragment(frag2);
      return frag;
   case NODE_GT:
      frag1 = convCreateLinearFragment(cset, node->left, timestat);
      convCheckInteger(frag1, node->left);
      frag2 = convCreateLinearFragment(cset, node->right, timestat);
      convCheckInteger(frag2, node->right);
      frag = convCreateGreaterFragment(cset, frag1, frag2, timestat);
      consDestroyFragment(frag1);
      consDestroyFragment(frag2);
      return frag;
   case NODE_LE:
      frag1 = convCreateLinearFragment(cset, node->left, timestat);
      convCheckInteger(frag1, node->left);
      frag2 = convCreateLinearFragment(cset, node->right, timestat);
      convCheckInteger(frag2, node->right);
      frag = convCreateLessequalFragment(cset, frag1, frag2, timestat);
      consDestroyFragment(frag1);
      consDestroyFragment(frag2);
      return frag;
   case NODE_GE:
      frag1 = convCreateLinearFragment(cset, node->left, timestat);
      convCheckInteger(frag1, node->left);
      frag2 = convCreateLinearFragment(cset, node->right, timestat);
      convCheckInteger(frag2, node->right);
      frag = convCreateGreaterequalFragment(cset, frag1, frag2, timestat);
      consDestroyFragment(frag1);
      consDestroyFragment(frag2);
      return frag;
   case NODE_UNION:
      break;
   case NODE_SETIN:
      break;
   case NODE_CASE:
      return convCreateCaseFragment(cset, node, timestat);
   default:
      errorMessage("invalid node type");
      abort();
   }

   return NULL;
}


/* CREATING CONSTRAINTS *****************************************************/

/**< Create constraint for an XOR.
 */
static void convCreateXorConstraint(
   CONSTRAINTSET*            cset,
   const CONSTRAINTFRAGMENT* frag1,
   const CONSTRAINTFRAGMENT* frag2,
   TIMESTAT                  timestat)
{
   CONSTRAINTFRAGMENT* frag;
   CONSTRAINT*         cons;

   /* add comment */
   consAddToSet(cset, consCreateOpComment("TOPLEVEL XOR", NULL, frag1, frag2, NULL));

   assert(consIsFragmentBoolean(frag1));
   assert(consIsFragmentBoolean(frag2));
   frag = consCopyFragment(frag1);
   consAddFragmentFragment(frag, frag2);
   cons = consCreateFromFragment(frag, INEQTYPE_EQUAL, 1);
   consAddToSet(cset, cons);
   consDestroyFragment(frag);
}

/**< Create constraints for an equality.
 *   This method produces constraints for a top-level equality
 *   <frag1> = <frag2> and adds them to the given constraint set.
 *   The two input fragments must have compatible types.
 */
static void convCreateEqualityConstraint(CONSTRAINTSET* cset,
					 const CONSTRAINTFRAGMENT* frag1,
					 const CONSTRAINTFRAGMENT* frag2)
{
   /* SMV CODE:
    *   <exp1> = <exp2>
    *
    * LP (for integer or binary expression types):
    *   <exp1> - <exp2> = 0
    *
    * LP (for enumeration types if <exp2> is constant):
    *   <exp1>[<exp2>] = 1
    *
    * LP (for enumeration types if both expressions are non-constant):
    *   <exp1>[atom] - <exp2>[atom] = 0,  for each atom in <exp1> \cap <exp2>
    *   <exp1>[atom] = 0,  for each atom in <exp1> \setminus <exp2>
    *   <exp2>[atom] = 0,  for each atom in <exp2> \setminus <exp1>
    *
    * TBD:
    * - Comparison of enumerations could do slightly better if intersection
    *   of types is found to be empty.
    * - Comparison of arrays?
    */

   CONSTRAINT* cons = NULL;
   CONSTRAINTFRAGMENT* frag = NULL;
   CONSTRAINTFRAGMENT* partfrag = NULL;
   const SYM* typesym1 = NULL;
   const SYM* typesym2 = NULL;
   SYM* typeunion = NULL;
   SYM* atom = NULL;
   SYMTABLE* symtable = NULL;
   Bool in1;
   Bool in2;

   assert(cset != NULL);
   assert(frag1 != NULL);
   assert(frag2 != NULL);
   assert(consGetFragmentType(frag1) == consGetFragmentType(frag2));

   /* add comment */
   consAddToSet(cset, consCreateOpComment("TOPLEVEL EQUALITY", NULL, frag1, frag2, NULL));

   switch (consGetFragmentType(frag1))
   {
   case FRAGTYPE_SCALAR:
      /* Constraint: <exp1> - <exp2> = 0 */
      frag = consCopyFragment(frag2);
      consMultiplyFragment(frag, -1);
      consAddFragmentFragment(frag, frag1);
      cons = consCreateFromFragment(frag, INEQTYPE_EQUAL, 0);
      consDestroyFragment(frag);
      consAddToSet(cset, cons);
      break;
   case FRAGTYPE_ENUM:
      typesym1 = consGetFragmentTypeSym(frag1);
      typesym2 = consGetFragmentTypeSym(frag2);
      if (consIsFragmentConst(frag1) && consIsFragmentConst(frag2)) {
	 if (typesym1 == typesym2) {
	    /* Nothing to be done for TRUE constraint */
	 } else {
	    /* Create the constraint 0 <= -1, which is always FALSE ... */
	    cons = consCreate(INEQTYPE_LESS, -1);
	    consAddToSet(cset, cons);
            printf("Infeasibility detected in enumeration equality.\n");
	 }
      } else if (!consIsFragmentConst(frag1) && consIsFragmentConst(frag2)) {
	 /* Constraint: <exp1>[<exp2>] = 1 */
	 frag = consCopyFragment(frag1);
	 consAppendFragmentEnumIndex(frag, typesym2);
	 cons = consCreateFromFragment(frag, INEQTYPE_EQUAL, 1);
	 consAddToSet(cset, cons);
	 consDestroyFragment(frag);
      } else if (consIsFragmentConst(frag1) && !consIsFragmentConst(frag2)) {
	 /* Constraint: <exp2>[<exp1>] = 1 */
	 frag = consCopyFragment(frag2);
	 consAppendFragmentEnumIndex(frag, typesym1);
	 cons = consCreateFromFragment(frag, INEQTYPE_EQUAL, 1);
	 consAddToSet(cset, cons);
	 consDestroyFragment(frag);
      } else {
	 typeunion = symCreateDeepCopy(typesym1);
	 symUniteEnum(typeunion, typesym2);
	 for (symtable = typeunion->value.enumvar.elemlist;
	      symtable != NULL;
	      symtable = symtable->next) {
	    atom = symtable->sym;
	    in1 = symContainsEnumAtom(typesym1, atom);
	    in2 = symContainsEnumAtom(typesym2, atom);
	    if (in1 && in2) {
	       /* Constraint: <exp1>[atom] - <exp2>[atom] = 0 */
	       frag = consCopyFragment(frag1);
	       consAppendFragmentEnumIndex(frag, atom);
	       partfrag = consCopyFragment(frag2); 
	       consAppendFragmentEnumIndex(partfrag, atom);
	       consMultiplyFragment(partfrag, -1);
	       consAddFragmentFragment(frag, partfrag);
	       consDestroyFragment(partfrag);
	       cons = consCreateFromFragment(frag, INEQTYPE_EQUAL, 0);
	       consAddToSet(cset, cons);
	       consDestroyFragment(frag);
	    } else if (in1) {
	       /* Constraint: <exp1>[atom] = 0 */
	       frag = consCopyFragment(frag1);
	       consAppendFragmentEnumIndex(frag, atom);
	       cons = consCreateFromFragment(frag, INEQTYPE_EQUAL, 0);
	       consAddToSet(cset, cons);
	       consDestroyFragment(frag);
	    } else if (in2) {
	       /* Constraint: <exp2>[atom] = 0 */
	       frag = consCopyFragment(frag2);
	       consAppendFragmentEnumIndex(frag, atom);
	       cons = consCreateFromFragment(frag, INEQTYPE_EQUAL, 0);
	       consAddToSet(cset, cons);
	       consDestroyFragment(frag);
	    } else {
	       assert(FALSE);  /* Atom from type union not found! */
	    }
	 }
	 symDestroy(typeunion);
      }
      break;
   default:
      assert(FALSE);  /* Unsupported equality! */
      break;
   }
}


/**< Create constraints for an inequality.
 *   This method produces constraints for a top-level inequality
 *   <frag1> != <frag2> and adds them to the given constraint set.
 *   The two input fragments must have compatible types.
 */
static void convCreateInequalityConstraint(
   CONSTRAINTSET* cset,
   const CONSTRAINTFRAGMENT* frag1,
   const CONSTRAINTFRAGMENT* frag2,
   TIMESTAT timestat)
{
   /* SMV CODE:
    *   <exp1> != <exp2>
    *
    * LP (for binary expression type):
    *   <exp1> xor <exp2>
    *
    * LP (for integer expression type):
    *   <<exp1> < <exp2>> + <<exp1> > <exp2>> = 1
    *
    * LP (for enumeration types if <exp2> is constant):
    *   <exp1>[<exp2>] = 0
    *
    * LP (for enumeration types if both expressions are non-constant):
    *   <exp1>[atom] + <exp2>[atom] <= 1,  for each atom in <exp1> \cap <exp2>
    *
    * TBD:
    * - Comparison of arrays?
    */

   CONSTRAINT* cons = NULL;
   CONSTRAINTFRAGMENT* frag = NULL;
   CONSTRAINTFRAGMENT* partfrag = NULL;
   const SYM* typesym1 = NULL;
   const SYM* typesym2 = NULL;
   SYM* typeintersection = NULL;
   SYM* atom = NULL;
   SYMTABLE* symtable = NULL;

   assert(cset != NULL);
   assert(frag1 != NULL);
   assert(frag2 != NULL);
   assert(consGetFragmentType(frag1) == consGetFragmentType(frag2));

   /* add comment */
   consAddToSet(cset, consCreateOpComment("TOPLEVEL INEQUALITY", NULL, frag1, frag2, NULL));

   switch (consGetFragmentType(frag1))
   {
   case FRAGTYPE_SCALAR:
      /* check, if both sides are binary */
      if( consIsFragmentBoolean(frag1) && consIsFragmentBoolean(frag2) )
      {
         /* Constraint: <exp1> xor <exp2> */
         convCreateXorConstraint(cset, frag1, frag2, timestat);
      }
      else
      {
         /* Constraint: <<exp1> < <exp2>> + <<exp1> > <exp2>> = 1 */
         frag = convCreateLessFragment(cset, frag1, frag2, timestat);
         partfrag = convCreateGreaterFragment(cset, frag1, frag2, timestat);
         consAddFragmentFragment(frag, partfrag);
         consDestroyFragment(partfrag);
         cons = consCreateFromFragment(frag, INEQTYPE_EQUAL, 1);
         consDestroyFragment(frag);
         consAddToSet(cset, cons);
      }
      break;
   case FRAGTYPE_ENUM:
      typesym1 = consGetFragmentTypeSym(frag1);
      typesym2 = consGetFragmentTypeSym(frag2);
      if (consIsFragmentConst(frag1) && consIsFragmentConst(frag2)) {
	 if (typesym1 != typesym2) {
	    /* Nothing to be done for TRUE constraint */
	 } else {
	    /* Create the constraint 0 <= -1, which is always FALSE ... */
	    cons = consCreate(INEQTYPE_LESS, -1);
	    consAddToSet(cset, cons);
            printf("Infeasibility detected in enumeration inequality.\n");
	 }
      } else if (!consIsFragmentConst(frag1) && consIsFragmentConst(frag2)) {
	 /* Constraint: <exp1>[<exp2>] = 0 */
	 frag = consCopyFragment(frag1);
	 consAppendFragmentEnumIndex(frag, typesym2);
	 cons = consCreateFromFragment(frag, INEQTYPE_EQUAL, 0);
	 consAddToSet(cset, cons);
	 consDestroyFragment(frag);
      } else if (consIsFragmentConst(frag1) && !consIsFragmentConst(frag2)) {
	 /* Constraint: <exp2>[<exp1>] = 0 */
	 frag = consCopyFragment(frag2);
	 consAppendFragmentEnumIndex(frag, typesym1);
	 cons = consCreateFromFragment(frag, INEQTYPE_EQUAL, 0);
	 consAddToSet(cset, cons);
	 consDestroyFragment(frag);
      } else {
	 typeintersection = symCreateDeepCopy(typesym1);
	 symIntersectEnum(typeintersection, typesym2);
	 for (symtable = typeintersection->value.enumvar.elemlist;
	      symtable != NULL;
	      symtable = symtable->next) {
	    atom = symtable->sym;
	    /* Constraint: <exp1>[atom] + <exp2>[atom] <= 1 */
	    frag = consCopyFragment(frag1);
	    consAppendFragmentEnumIndex(frag, atom);
	    partfrag = consCopyFragment(frag2); 
	    consAppendFragmentEnumIndex(partfrag, atom);
	    consAddFragmentFragment(frag, partfrag);
	    consDestroyFragment(partfrag);
	    cons = consCreateFromFragment(frag, INEQTYPE_LESS, 1);
	    consAddToSet(cset, cons);
	    consDestroyFragment(frag);
	 }
	 symDestroy(typeintersection);
      }
      break;
   default:
      assert(FALSE);  /* Unsupported inequality! */
      break;
   }
}


/**< Creates inequalities for a case constraint on top level.
 */
static void convCreateCaseConstraint(
   CONSTRAINTSET* cset,
   const Node*    node,
   TIMESTAT       timestat)
{
   /* SMV CODE:
    *   case
    *      <cond1> : <then1>;
    *      <cond2> : <then2>;
    *      <cond3> : <then3>;
    *      ...
    *      <condN> : <thenN>;
    *    esac;
    *
    * PARSE TREE STRUCTURE:
    *   NODE_CASE
    *   +- NODE_COLON
    *   |  +- <cond1>  (must have boolean result type)
    *   |  +- <then1>  (must have boolean result type)
    *   +- NODE_CASE
    *     +- NODE_COLON
    *     |  +- <cond2>  (must have boolean result type)
    *     |  +- <then2>  (must have boolean result type)
    *     +- NODE_CASE
    *        ...
    *        ...
    *        +- NODE_TRUEEXP
    */

   const Node* left;
   const Node* right;
   CONSTRAINT* cons;
   CONSTRAINTFRAGMENT* truecondition;
   CONSTRAINTFRAGMENT* condition;
   CONSTRAINTFRAGMENT* notcondition;
   CONSTRAINTFRAGMENT* thiscondition;
   CONSTRAINTFRAGMENT* thenfragment;
   Bool alwaystrue;

   assert(cset != NULL);
   assert(node != NULL);
   assert(node->type == NODE_CASE);

   /* add comment */
   consAddToSet(cset, consCreateOpComment("TOPLEVEL CASE", NULL, NULL, NULL, NULL));

   truecondition = consCreateFragmentConst(0);
   alwaystrue = FALSE;

   while( !alwaystrue && node->type != NODE_TRUEEXP )
   {
      left = node->left;
      assert(left != NULL);
      assert(left->type == NODE_COLON);
      assert(left->left != NULL);
      assert(left->right != NULL);
      
      right = node->right;
      assert(right != NULL);
      assert(right->type == NODE_CASE || right->type == NODE_TRUEEXP);

      /* get fragments for <cond>:<then> */
      thiscondition = convCreateLinearFragment(cset, left->left, timestat);
      thenfragment = convCreateLinearFragment(cset, left->right, timestat);

      /* check types of fragments */
      convCheckBoolean(thiscondition, left->left);
      convCheckBoolean(thenfragment, left->right);

      /* check, if <cond> can be true */
      if( consGetFragmentMaxVal(thiscondition) >= 1 )
      {
         /* construct fragment (1-<cond_i>) */
         notcondition = consCopyFragment(thiscondition);
         consNegateFragment(notcondition);
         
         /* construct fragment <cond_1> + <cond_2> + ... + <cond_{i-1}> + (1-<cond_i>) + <then> */
         condition = consCopyFragment(truecondition);
         consAddFragmentFragment(condition, notcondition);
         consAddFragmentFragment(condition, thenfragment);
         
         /* create constraint: <cond_1> + <cond_2> + ... + <cond_{i-1}> + (1-<cond_i>) + <then> >= 1 */
         cons = consCreateFromFragment(condition, INEQTYPE_GREATER, 1);
         consAddToSet(cset, cons);
         
         /* update truecondition */
         consAddFragmentFragment(truecondition, thiscondition);
      
         /* free temporary fragments */
         consDestroyFragment(condition);
         consDestroyFragment(notcondition);
      }

      /* check, if last <cond> is always true */
      alwaystrue = ( consGetFragmentMinVal(thiscondition) >= 1 );

      /* free temporary fragments */
      consDestroyFragment(thiscondition);
      consDestroyFragment(thenfragment);

      node = right;
   }

   /* free truecondition */
   consDestroyFragment(truecondition);
}

static void convCreateConstraints(CONSTRAINTSET* cset,
				  const Node* node,
				  TIMESTAT timestat)
{
   CONSTRAINT* cons = NULL;
   CONSTRAINTFRAGMENT* frag = NULL;
   CONSTRAINTFRAGMENT* frag1 = NULL;
   CONSTRAINTFRAGMENT* frag2 = NULL;
   Node* left = NULL;
   char errmsg[1024];

   assert(cset != NULL);
   
   if( node == NULL )
      return;
   
   consAddToSet(cset, consCreateLineComment(node));

   switch(node->type)
   {
   case NODE_TRUEEXP:
      /* Nothing to be done for TRUE ... */
      break;
   case NODE_FALSEEXP:
      /* Create the constraint 0 <= -1, which is always FALSE ... */
      consAddToSet(cset, consCreateComment("  infeasibility detected"));
      cons = consCreate(INEQTYPE_LESS, -1);
      consAddToSet(cset, cons);
      printf("Infeasibility detected: FALSE in line %d.\n", node->lineno);
      break;
   case NODE_CONS: 
   case NODE_AND: 
      convCreateConstraints(cset, node->left, timestat);
      convCreateConstraints(cset, node->right, timestat);
      break;
   case NODE_ATOM: 
   case NODE_ARRAY: 
      frag = convCreateLinearFragment(cset, node, timestat);
      convCheckBoolean(frag, node->left);
      cons = consCreateFromFragment(frag, INEQTYPE_EQUAL, 1);
      consAddToSet(cset, cons);
      consDestroyFragment(frag);
      break;
   case NODE_NEXT:
      if (timestat == TIMESTAT_ACTUAL) {
	 convCreateConstraints(cset, node->left, TIMESTAT_NEXT);
      } else {
	 sprintf(errmsg, "Nested next() encountered in line %d", node->lineno);
	 errorMessage(errmsg);
	 abort();
      }
      break;
   case NODE_NOT:
      frag = convCreateLinearFragment(cset, node->left, timestat);
      convCheckBoolean(frag, node->left);
      cons = consCreateFromFragment(frag, INEQTYPE_EQUAL, 0);
      consAddToSet(cset, cons);
      consDestroyFragment(frag);
      break;
   case NODE_OR:
      /* TBD: flatten nested OR nodes */
      frag1 = convCreateLinearFragment(cset, node->left, timestat);
      convCheckBoolean(frag1, node->left);
      frag2 = convCreateLinearFragment(cset, node->right, timestat);
      convCheckBoolean(frag2, node->right);
      consAddFragmentFragment(frag1, frag2);
      cons = consCreateFromFragment(frag1, INEQTYPE_GREATER, 1);
      consAddToSet(cset, cons);
      consDestroyFragment(frag2);
      consDestroyFragment(frag1);
      break;      
   case NODE_IMPLIES: 
      frag1 = convCreateLinearFragment(cset, node->left, timestat);
      convCheckBoolean(frag1, node->left);
      frag2 = convCreateLinearFragment(cset, node->right, timestat);
      convCheckBoolean(frag2, node->right);
      consMultiplyFragment(frag2, -1);
      consAddFragmentFragment(frag1, frag2);
      cons = consCreateFromFragment(frag1, INEQTYPE_LESS, 0);
      consAddToSet(cset, cons);
      consDestroyFragment(frag2);
      consDestroyFragment(frag1);
      break;      
   case NODE_XOR: 
      frag1 = convCreateLinearFragment(cset, node->left, timestat);
      convCheckBoolean(frag1, node->left);
      frag2 = convCreateLinearFragment(cset, node->right, timestat);
      convCheckBoolean(frag2, node->right);
      convCreateXorConstraint(cset, frag1, frag2, timestat);
      consDestroyFragment(frag2);
      consDestroyFragment(frag1);
      break;      
   case NODE_XNOR:  /* ??? CHECK: XNOR == IFF */
   case NODE_IFF: 
      frag1 = convCreateLinearFragment(cset, node->left, timestat);
      convCheckBoolean(frag1, node->left);
      frag2 = convCreateLinearFragment(cset, node->right, timestat);
      convCheckBoolean(frag2, node->right);
      consMultiplyFragment(frag2, -1);
      consAddFragmentFragment(frag1, frag2);
      cons = consCreateFromFragment(frag1, INEQTYPE_EQUAL, 0);
      consAddToSet(cset, cons);
      consDestroyFragment(frag2);
      consDestroyFragment(frag1);
      break;      
   case NODE_EQDEF:
      left = node->left;
      if (left->type == NODE_SMALLINIT) {
	 if (timestat != TIMESTAT_ACTUAL) {
	    sprintf(errmsg, "Unexpected init() in next()/init() context in line %d", left->lineno);
	    errorMessage(errmsg);
	    abort();
	 }
	 timestat = TIMESTAT_INIT;
         assert(left->right == NULL);
	 left = left->left;
      }
      frag1 = convCreateLinearFragment(cset, left, timestat);
      frag2 = convCreateLinearFragment(cset, node->right, timestat);
      convCheckAssignmentTypes(frag1, frag2, left);
      convCreateEqualityConstraint(cset, frag1, frag2);
      consDestroyFragment(frag1);
      consDestroyFragment(frag2);
      break;      
   case NODE_EQUAL: 
      frag1 = convCreateLinearFragment(cset, node->left, timestat);
      frag2 = convCreateLinearFragment(cset, node->right, timestat);
      convCheckEqualityTypes(frag1, frag2, node);
      convCreateEqualityConstraint(cset, frag1, frag2);
      consDestroyFragment(frag1);
      consDestroyFragment(frag2);
      break;
   case NODE_NOTEQUAL: 
      frag1 = convCreateLinearFragment(cset, node->left, timestat);
      frag2 = convCreateLinearFragment(cset, node->right, timestat);
      convCheckEqualityTypes(frag1, frag2, node->left);
      convCreateInequalityConstraint(cset, frag1, frag2, timestat);
      consDestroyFragment(frag1);
      consDestroyFragment(frag2);
      break;
   case NODE_LT: 
      frag1 = convCreateLinearFragment(cset, node->left, timestat);
      frag2 = convCreateLinearFragment(cset, node->right, timestat);
      convCheckInteger(frag1, node->left);
      convCheckInteger(frag2, node->right);
      consMultiplyFragment(frag2, -1);
      consAddFragmentFragment(frag1, frag2);
      cons = consCreateFromFragment(frag1, INEQTYPE_LESS, -1);
      consAddToSet(cset, cons);      
      consDestroyFragment(frag2);
      consDestroyFragment(frag1);
      break;
   case NODE_GT: 
      frag1 = convCreateLinearFragment(cset, node->left, timestat);
      frag2 = convCreateLinearFragment(cset, node->right, timestat);
      convCheckInteger(frag1, node->left);
      convCheckInteger(frag2, node->right);
      consMultiplyFragment(frag2, -1);
      consAddFragmentFragment(frag1, frag2);
      cons = consCreateFromFragment(frag1, INEQTYPE_GREATER, 1);
      consAddToSet(cset, cons);      
      consDestroyFragment(frag2);
      consDestroyFragment(frag1);
      break;
   case NODE_LE: 
      frag1 = convCreateLinearFragment(cset, node->left, timestat);
      frag2 = convCreateLinearFragment(cset, node->right, timestat);
      convCheckInteger(frag1, node->left);
      convCheckInteger(frag2, node->right);
      consMultiplyFragment(frag2, -1);
      consAddFragmentFragment(frag1, frag2);
      cons = consCreateFromFragment(frag1, INEQTYPE_LESS, 0);
      consAddToSet(cset, cons);      
      consDestroyFragment(frag2);
      consDestroyFragment(frag1);
      break;
   case NODE_GE: 
      frag1 = convCreateLinearFragment(cset, node->left, timestat);
      frag2 = convCreateLinearFragment(cset, node->right, timestat);
      convCheckInteger(frag1, node->left);
      convCheckInteger(frag2, node->right);
      consMultiplyFragment(frag2, -1);
      consAddFragmentFragment(frag1, frag2);
      cons = consCreateFromFragment(frag1, INEQTYPE_GREATER, 0);
      consAddToSet(cset, cons);      
      consDestroyFragment(frag2);
      consDestroyFragment(frag1);
      break;
   case NODE_SETIN: 
      errorMessage("SETIN not yet implemented");
      abort();
   case NODE_CASE:
      convCreateCaseConstraint(cset, node, timestat);
      break;
   case NODE_PROPERTY:
      frag = convCreateLinearFragment(cset, node->left, timestat);
      convCheckBoolean(frag, node->left);
      consAddToSet(cset, consCreateComment("PROPERTY"));
      cons = consCreateProperty(frag);
      consAddToSet(cset, cons);
      consDestroyFragment(frag);
      break;
   default:
      errorMessage("invalid node type");
      abort();
   }
}


static void convSubtree(CONSTRAINTSET* cset, const Node* node)
{
   Node* tmpnode;

   assert(cset != NULL);

   if( node == NULL )
      return;

   consAddToSet(cset, consCreateLineComment(node));

   switch(node->type)
   {
   case NODE_ERROR: 
      errorMessage("error in node");
      abort();
   case NODE_FALSEEXP: 
   case NODE_TRUEEXP: 
   case NODE_NUMBER: 
   case NODE_ATOM: 
   case NODE_DOT: 
   case NODE_GOTO: 
   case NODE_LET: 
   case NODE_EVAL: 
      break;
   case NODE_INIT:  /* INIT statement */
      consAddToSet(cset, consCreateComment("INIT block"));
      convCreateConstraints(cset, node->left, TIMESTAT_INIT);
      break;
   case NODE_FAILNESS: 
      break;
   case NODE_TRANS: /* TRANS statement */
      consAddToSet(cset, consCreateComment("TRANS block"));
      convCreateConstraints(cset, node->left, TIMESTAT_ACTUAL);
      break;
   case NODE_CONSTRAINT: 
   case NODE_SIMPWFF: 
   case NODE_CONTEXT: 
   case NODE_CTLWFF: 
   case NODE_LTWFF: 
   case NODE_EQDEF: 
      break;
   case NODE_CONS: 
      convSubtree(cset, node->left);
      convSubtree(cset, node->right);
      break;
   case NODE_VAR: 
   case NODE_IVAR:
   case NODE_COLON: 
   case NODE_BOOLEAN: 
   case NODE_SCALAR: 
   case NODE_ARRAY: 
   case NODE_PROCESS: 
   case NODE_TWODOTS: 
   case NODE_UMINUS: 
   case NODE_MODTYPE: 
      break;
   case NODE_ASSIGN: /* ASSIGN statement */
      consAddToSet(cset, consCreateComment("ASSIGN block"));
      convCreateConstraints(cset, node->left, TIMESTAT_ACTUAL);
      break;
   case NODE_AND: 
   case NODE_SMALLINIT: 
   case NODE_NEXT: 
      break;
   case NODE_INVAR:  /* INVAR statement */
      consAddToSet(cset, consCreateComment("INVAR block"));
      convCreateConstraints(cset, node->left, TIMESTAT_ACTUAL);
      break;
   case NODE_DEFINE: /* DEFINE statement */
      break;
   case NODE_JUSTICE: 
   case NODE_COMPASSION: 
   case NODE_SPEC: 
   case NODE_LTLSPEC: 
      break;
   case NODE_INVARSPEC: /* INVARSPEC statement */
      consAddToSet(cset, consCreateComment("INVARSPEC block"));
      tmpnode = node_new(NODE_PROPERTY, node->left, NULL, node->lineno, node->column);
      assert(tmpnode != NULL);
      convCreateConstraints(cset, tmpnode, TIMESTAT_ACTUAL);
      tmpnode->left = NULL;
      node_free(&tmpnode);
      break;
   case NODE_COMPUTE: 
   case NODE_MINU: 
   case NODE_MAXU: 
   case NODE_SELF: 
   case NODE_IMPLIES: 
   case NODE_IFF: 
   case NODE_OR: 
   case NODE_XOR: 
   case NODE_XNOR: 
   case NODE_NOT: 
   case NODE_PLUS: 
   case NODE_MINUS: 
   case NODE_TIMES: 
   case NODE_IMPLEMENTS: 
   case NODE_INPUT: 
   case NODE_OUTPUT: 
   case NODE_FAIRNESS: 
   case NODE_DIVIDE: 
   case NODE_MOD: 
   case NODE_EQUAL: 
   case NODE_NOTEQUAL: 
   case NODE_LT: 
   case NODE_GT: 
   case NODE_LE: 
   case NODE_GE: 
   case NODE_UNION: 
   case NODE_SETIN: 
   case NODE_CASE: 
   case NODE_EX: 
   case NODE_AX: 
   case NODE_EF: 
   case NODE_AF: 
   case NODE_EG: 
   case NODE_AG: 
   case NODE_AU: 
   case NODE_EU: 
   case NODE_EBF: 
   case NODE_ABF: 
   case NODE_EBG: 
   case NODE_ABG: 
   case NODE_UNTIL: 
   case NODE_SINCE: 
   case NODE_OP_NEXT: 
   case NODE_OP_PREC: 
   case NODE_OP_NOTPRECNOT: 
   case NODE_OP_GLOBAL: 
   case NODE_OP_HISTORICAL: 
   case NODE_OP_FUTURE: 
   case NODE_OP_ONCE: 
   case NODE_LTLWFF: 
   case NODE_COMPWFF: 
      break;
   case NODE_MODULE: 
      assert(node->left != NULL && node->left->type == NODE_MODTYPE);
      convSubtree(cset, node->right);
      break;
   case NODE_ABU: 
   case NODE_EBU:
   case NODE_ISA:
   case NODE_SIGMA:
      break;
   default:
      errorMessage("invalid node type");
      abort();
   }
}

/* MAIN ROUTINE *************************************************************/

CONSTRAINTSET* convTreeToCSet(const Node* root)
{
   CONSTRAINTSET* cset;
   SYMTABLE*      symtable;
   Node*          normroot;

   /* ??? the tree should be transformed in the following way:
      - all next-expressions should be innermost directly in font of variable
        identifiers
      - substitute  "x in case <c1>:<e1>; <c2>:<e2>; ... esac;" by
        "case <c1>: x in <e1>; <c2>: x in <e2>; ... esac;"
      - eliminate DOT (".") operators by catenating symbol names
      - all defines are replaced by their subtree

      - evaluate constants as much as possible
      - recognize the constant FALSE occurring on top-level and stop
      - move negations inward

      - ?subexpressions which occur multiply should be recognized?

      normroot = normalizeTree(root);
   */
   normroot = (Node*) root; /* ??? */

   symtable = symCreateTable(normroot);
   cset = consCreateCSet(symtable);

   /* convert the tree into constraints */
   convSubtree(cset, normroot);

   /* create enum equalities */
   consCreateEnumEqualities(cset);

   /* assign MIP variable numbers */
   consAssignVariableNumbers(cset);

   /* ??? nodeFree(normroot); */

   return cset;
}
