%{
/**CFile***********************************************************************

  FileName    [grammar.y]

  PackageName [parser]

  Synopsis    [Yacc for NuSMV input language parser]

  SeeAlso     [input.lex]

  Author      [Marco Roveri]

  Copyright   [
  This file is part of the ``parser'' package of NuSMV version 2. 
  Copyright (C) 1998-2001 by CMU and ITC-irst. 

  NuSMV version 2 is free software; you can redistribute it and/or 
  modify it under the terms of the GNU Lesser General Public 
  License as published by the Free Software Foundation; either 
  version 2 of the License, or (at your option) any later version.

  NuSMV version 2 is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of 
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public 
  License along with this library; if not, write to the Free Software 
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA.

  For more information of NuSMV see <http://nusmv.irst.itc.it>
  or email to <nusmv-users@irst.itc.it>.
  Please report bugs to <nusmv-users@irst.itc.it>.

  To contact the NuSMV development board, email to <nusmv@irst.itc.it>. ]

******************************************************************************/

#include "malloc.h"
#include "node.h"

#define YYERROR_VERBOSE 1
   
extern void yyerror(const char* s);
extern void set_parse_tree(Node* pt);

/* 
node_ptr parse_tree;
node_ptr parse_tree_int;
int parse_command_flag;
*/

%}
%pure_parser

/*
  The number of conflicts: shift/reduce expected.
  If more, than a warning message is printed out.
*/
%expect 1

%union {
  Node* node;
}

/*
  All of the terminal grammar symbols (tokens recognized by the lexical analyzer) 
  Note: all binary operators associate from left to right,  operators are
        listed from lowest to highest priority 

  Note: The following token are not used inside the grammar, but are
  used by other modules inside the system (i.e. the compiler, mc).
  STEP RESET ASYNC MODTYPE LAMBDA CONTEXT EU AU EBU ABU MINU MAXU
  FORMAT CONSTANT SCALAR CONS OVER BDD ATLINE APROPOS IFTHENELSE
  QUOTE DL_ATOM APATH EPATH 
*/

%left MODULE PROCESS 
%left VAR IVAR DEFINE INIT TRANS INVAR SPEC LTLSPEC COMPUTE 
%left INVARSPEC FAIRNESS COMPASSION JUSTICE
%left ISA ASSIGN INPUT OUTPUT IMPLEMENTS
%left BOOLEAN ARRAY OF 
%left SEMI LP RP LB RB LCB RCB
%left EQDEF TWODOTS 
%left <node> FALSEEXP TRUEEXP
%left SELF SIGMA
%left CASE ESAC COLON

%left <node> ATOM
%left <node> NUMBER

%left  COMMA
%right IMPLIES
%left  IFF
%left  OR XOR XNOR
%left  AND
%left  NOT
%left  EX AX EF AF EG AG EE AA SINCE UNTIL TRIGGERED RELEASES EBF EBG ABF ABG BUNTIL MMIN MMAX
%left  OP_NEXT OP_GLOBAL OP_FUTURE
%left  OP_PREC OP_NOTPRECNOT OP_HISTORICAL OP_ONCE
%left  EQUAL NOTEQUAL LT GT LE GE
%left  UNION
%left  SETIN
%left  MOD
%left  PLUS MINUS
%left  TIMES DIVIDE
%left  UMINUS		/* supplies precedence for unary minus */
%left  NEXT SMALLINIT
%left  DOT

/* all nonterminals return a parse tree node */
%type <node> module_list module declarations declaration var input_var var_list
%type <node> ivar_list type itype isa init trans invar define define_list spec
%type <node> ltlspec compute fairness justice compassion invarspec 
%type <node> atom_list decl_var_id var_id subrange number simple_expr 
%type <node> assign_expr init_expr next_expr justice_expr compassion_expr 
%type <node> fair_expr invar_expr trans_expr ctl_expr compute_expr s_case_list 
%type <node> n_case_list constant ltl_expr ltl_orexpr ltl_andexpr define_expr
%type <node> ltl_binary_expr ltl_atomexpr ltl_case_expr ltl_simple_expr 
%type <node> module_sign module_type simple_expr_list assign assign_list
%type <node> assign_type simple_expr_set next_expr_set ctl_expr_set 
%type <node> ltl_expr_set input output implements term_list
%type <node> constant_list actual_params 
%type <node> ctl_case_list

/* %type <node> trace state */

/* currently not used: SEE GRAMMAR RULES 
%type next_expr_list
*/

%start begin
%%
begin         : module_list { set_parse_tree($1); }
              ;

/*
 An NuSMV program is a repetition of modules.
 Each module has a signature and a body.
*/
module_list   : module {$$ = cons($1, NODE_NIL);}
              | module_list module {$$ = cons($2, $1);}
              ;
module        : MODULE module_sign declarations {$$ = new_node(NODE_MODULE, $2, $3);}
              ;
module_sign   : ATOM {$$ = new_node(NODE_MODTYPE, $1, NODE_NIL);}
              | ATOM LP RP {$$ = new_node(NODE_MODTYPE, $1, NODE_NIL);}
              | ATOM LP atom_list RP {$$ = new_node(NODE_MODTYPE, $1, $3);}
              ;
atom_list     : ATOM {$$ = cons($1, NODE_NIL);}
              | atom_list COMMA ATOM {$$ = cons($3, $1);}
              ;


/* The body of a module */
declarations  : {$$ = NODE_NIL;}
              | declarations declaration {$$ = cons($2, $1);}
              ;
declaration   : isa        
              | var
              | input_var
              | assign 
              | init
              | invar
              | trans
              | define
              | fairness  /* ignore */
              | justice   /* ignore */
              | compassion /* ignore */
              | invarspec
              | spec        /* later */ 
              | ltlspec     /* later */
              | compute     /* ignore */
              | implements  /* ignore */
              | input       /* ignore */
              | output      /* ignore */
              ;

/* Module macro-expansion */
isa           : ISA ATOM {$$ = new_node(NODE_ISA, $2, NODE_NIL);}
              ;

/* parse an optional semicolon */
optsemi      : | SEMI;

/*
 Variable declarations:
 This includes also the instantiation of module
 (in synchronous and asynchronous product).
*/
var           : VAR var_list {$$ = new_node(NODE_VAR, $2, NODE_NIL);}
              ;
input_var     : IVAR ivar_list {$$ = new_node(NODE_IVAR, $2, NODE_NIL);}
              ;
var_list      : {$$ = NODE_NIL;}
              | var_list decl_var_id COLON type SEMI {$$ = cons(new_node(NODE_COLON, $2, $4), $1);}
              ;
ivar_list     : {$$ = NODE_NIL;}
              | ivar_list decl_var_id COLON itype SEMI {$$ = cons(new_node(NODE_COLON, $2, $4), $1);}
              ;
type          : BOOLEAN {$$ = new_node(NODE_BOOLEAN, NODE_NIL, NODE_NIL);}
              | subrange
              | LCB constant_list RCB {$$ = new_node(NODE_SCALAR, $2, NODE_NIL);}
              | ARRAY subrange OF type {$$ = new_node(NODE_ARRAY, $2, $4);}
              | module_type
              | PROCESS module_type {$$ = new_node(NODE_PROCESS, $2, NODE_NIL);}
              ;
itype         : BOOLEAN {$$ = new_node(NODE_BOOLEAN, NODE_NIL, NODE_NIL);}
              | subrange
              | LCB constant_list RCB {$$ = new_node(NODE_SCALAR, $2, NODE_NIL);}
              | ARRAY subrange OF itype {$$ = new_node(NODE_ARRAY, $2, $4);}
              ;
constant_list : constant {$$ = cons($1, NODE_NIL);}
              | constant_list COMMA constant {$$ = cons($3, $1);}
              ;
constant      : ATOM
              | number
	      | FALSEEXP
	      | TRUEEXP
              ;
subrange      : number TWODOTS number {$$ = new_node(NODE_TWODOTS, $1, $3);}
              ;
number        : NUMBER
	      | PLUS NUMBER %prec UMINUS { $$ = $2; }
	      | MINUS NUMBER %prec UMINUS { $$ = new_node(NODE_UMINUS, $2, NODE_NIL); }
              ;
module_type   : ATOM {$$ = new_node(NODE_MODTYPE, $1, NODE_NIL);}
              | ATOM LP RP {$$ = new_node(NODE_MODTYPE, $1, NODE_NIL);}
              | ATOM LP actual_params RP {$$ = new_node(NODE_MODTYPE, $1, $3);}
              ;
actual_params : simple_expr_list
              ;

/* Assignments of initial, current or next value of variables */
assign        : ASSIGN assign_list {$$ = new_node(NODE_ASSIGN, $2, NODE_NIL);}
              ;
assign_list   : {$$ = NODE_NIL;}
              | assign_list assign_type {$$ = new_node(NODE_AND, $1, $2);}
              ;
assign_type   : var_id EQDEF simple_expr SEMI
                 {$$ = new_node(NODE_EQDEF, $1, $3);} 
              | SMALLINIT LP var_id RP EQDEF simple_expr SEMI
                { $$ = new_node(NODE_EQDEF, new_node(NODE_SMALLINIT, $3, NODE_NIL), $6);}
              | NEXT LP var_id RP EQDEF assign_expr SEMI
                { $$ = new_node(NODE_EQDEF, new_node(NODE_NEXT, $3, NODE_NIL), $6);}
              ;
assign_expr   : next_expr
              ;

/* Direct finite state machine definition (init, invar, trans) */
init          : INIT init_expr optsemi {$$ = new_node(NODE_INIT, $2, NODE_NIL);}
              ;
init_expr     : simple_expr
              ;
invar         : INVAR invar_expr optsemi {$$ = new_node(NODE_INVAR, $2, NODE_NIL);}
              ;
invar_expr    : simple_expr
              ;
trans         : TRANS trans_expr optsemi {$$ = new_node(NODE_TRANS, $2, NODE_NIL);}
              ;
trans_expr    : next_expr
              ;

/* Definitions */
define        : DEFINE define_list {$$ = new_node(NODE_DEFINE, $2, NODE_NIL);}
              ;
define_list   : {$$ = NODE_NIL;}
              | define_list decl_var_id EQDEF define_expr SEMI {$$ = cons(new_node(NODE_EQDEF, $2, $4), $1);}
              ;
define_expr   : next_expr
              ;

/* Fairness declarations */
fairness      : FAIRNESS fair_expr optsemi {$$ = new_node(NODE_JUSTICE, $2, NODE_NIL);}
              ;
fair_expr     : simple_expr
              ;

justice       : JUSTICE  justice_expr optsemi {$$ = new_node(NODE_JUSTICE, $2, NODE_NIL);}
              ;
justice_expr  : simple_expr
              ;

compassion    : COMPASSION compassion_expr optsemi {$$ = new_node(NODE_COMPASSION, $2, NODE_NIL);}
              ;
compassion_expr: LP simple_expr COMMA simple_expr RP {$$ = cons($2,$4);} 
               ;

/* Specifications and computation of min and max distance */
spec          : SPEC ctl_expr optsemi {$$ = new_node(NODE_SPEC, $2, NODE_NIL);}
              ;

ltlspec       : LTLSPEC ltl_expr optsemi {$$ = new_node(NODE_LTLSPEC, $2, NODE_NIL);}
              ;

invarspec     : INVARSPEC invar_expr optsemi {$$ = new_node(NODE_INVARSPEC, $2, NODE_NIL);}
              ;
 
compute       : COMPUTE compute_expr optsemi {$$ = new_node(NODE_COMPUTE, $2, NODE_NIL);}
              ;
compute_expr  : MMIN LB ctl_expr COMMA ctl_expr RB { $$ = new_node(NODE_MINU, $3, $5); }
              | MMAX LB ctl_expr COMMA ctl_expr RB { $$ = new_node(NODE_MAXU, $3, $5); }
              ;

/* Variable identifiers.
   decl_var_id is used for declarations; self not allowed.
   var_id is used to reference variables, includes self.
 */

decl_var_id   : ATOM
              | decl_var_id DOT ATOM {$$ = new_node(NODE_DOT, $1, $3);}
              | decl_var_id DOT NUMBER {$$ = new_node(NODE_DOT, $1, $3);}
              | decl_var_id LB simple_expr RB {$$ = new_node(NODE_ARRAY, $1, $3);}
              ;
var_id        : ATOM
              | SELF {$$ = new_node(NODE_SELF,NODE_NIL,NODE_NIL);}
              | var_id DOT ATOM {$$ = new_node(NODE_DOT, $1, $3);}
              | var_id DOT NUMBER {$$ = new_node(NODE_DOT, $1, $3);}
              | var_id LB simple_expr RB {$$ = new_node(NODE_ARRAY, $1, $3);}
              ;


/* Simple expressions. Do not involve next state variables or CLT. */

simple_expr_list : simple_expr {$$ = cons($1,NODE_NIL);}
	         | simple_expr_list COMMA simple_expr {$$ = cons($3, $1);}
                 ;
simple_expr   : var_id
              | number
              | subrange
              | FALSEEXP
              | TRUEEXP
              | LP simple_expr RP { $$ = $2; }
	      | simple_expr IMPLIES simple_expr { $$ = new_node(NODE_IMPLIES, $1, $3); }
	      | simple_expr IFF simple_expr { $$ = new_node(NODE_IFF, $1, $3); }
	      | simple_expr OR simple_expr { $$ = new_node(NODE_OR,$1, $3); }
              | simple_expr XOR simple_expr { $$ = new_node(NODE_XOR,$1, $3); }
              | simple_expr XNOR simple_expr { $$ = new_node(NODE_XNOR,$1, $3); }
	      | simple_expr AND simple_expr { $$ = new_node(NODE_AND, $1, $3); }
	      | NOT simple_expr { $$ = new_node(NODE_NOT, $2, NODE_NIL); }
              | simple_expr PLUS simple_expr { $$ = new_node(NODE_PLUS, $1, $3); }
              | simple_expr MINUS simple_expr { $$ = new_node(NODE_MINUS, $1, $3); }
              | simple_expr TIMES simple_expr { $$ = new_node(NODE_TIMES, $1, $3); }
              | simple_expr DIVIDE simple_expr { $$ = new_node(NODE_DIVIDE, $1, $3); }
              | simple_expr MOD simple_expr { $$ = new_node(NODE_MOD, $1, $3); }
              | simple_expr EQUAL simple_expr { $$ = new_node(NODE_EQUAL, $1, $3); }
              | simple_expr NOTEQUAL simple_expr { $$ = new_node(NODE_NOTEQUAL, $1, $3); }
              | simple_expr LT simple_expr { $$ = new_node(NODE_LT, $1, $3); }
              | simple_expr GT simple_expr { $$ = new_node(NODE_GT, $1, $3); }
              | simple_expr LE simple_expr { $$ = new_node(NODE_LE, $1, $3); }
              | simple_expr GE simple_expr { $$ = new_node(NODE_GE, $1, $3); }
              | LCB simple_expr_set RCB { $$ = $2; } 
              | simple_expr UNION simple_expr { $$ = new_node(NODE_UNION, $1, $3); }
              | simple_expr SETIN simple_expr { $$ = new_node(NODE_SETIN, $1, $3); }
              | CASE s_case_list ESAC { $$ = $2; }
              | SIGMA LB ATOM EQUAL subrange RB simple_expr
                { $$ = new_node(NODE_SIGMA, new_node(NODE_EQUAL, $3, $5), $7); }
	      ;
simple_expr_set : simple_expr {$$ = $1;}
                | simple_expr_set COMMA simple_expr {$$ = new_node(NODE_UNION, $1, $3);}
                ;
s_case_list   : {$$=new_node(NODE_TRUEEXP, NODE_NIL, NODE_NIL);}
              | simple_expr COLON simple_expr SEMI s_case_list
                {
	          $$ = new_node(NODE_CASE, new_node(NODE_COLON, $1, $3), $5);
	        }
              ;



/* Next expressions. Contain next state variables, but no CTL. */
next_expr     : var_id
              | number
              | subrange
              | FALSEEXP
              | TRUEEXP
/*              | NEXT LP var_id RP { $$ = new_node(NODE_NEXT, $3, NODE_NIL); } */
/* We dont want expression of this kind: next(alpha & next(beta))   */
              | NEXT LP simple_expr RP { $$ = new_node(NODE_NEXT, $3, NODE_NIL); } 
              | LP next_expr RP { $$ = $2; }
	      | next_expr IMPLIES next_expr { $$ = new_node(NODE_IMPLIES, $1, $3); }
	      | next_expr IFF next_expr { $$ = new_node(NODE_IFF, $1, $3); }
	      | next_expr OR next_expr { $$ = new_node(NODE_OR, $1, $3); }
	      | next_expr XOR next_expr { $$ = new_node(NODE_XOR, $1, $3); }
	      | next_expr XNOR next_expr { $$ = new_node(NODE_XNOR, $1, $3); }
	      | next_expr AND next_expr { $$ = new_node(NODE_AND, $1, $3); }
	      | NOT next_expr { $$ = new_node(NODE_NOT, $2, NODE_NIL); }
              | next_expr PLUS next_expr { $$ = new_node(NODE_PLUS, $1, $3); }
              | next_expr MINUS next_expr { $$ = new_node(NODE_MINUS, $1, $3); }
              | next_expr TIMES next_expr { $$ = new_node(NODE_TIMES, $1, $3); }
              | next_expr DIVIDE next_expr { $$ = new_node(NODE_DIVIDE, $1, $3); }
              | next_expr MOD next_expr { $$ = new_node(NODE_MOD, $1, $3); }
              | next_expr EQUAL next_expr { $$ = new_node(NODE_EQUAL, $1, $3); }
              | next_expr NOTEQUAL next_expr { $$ = new_node(NODE_NOTEQUAL, $1, $3); }
              | next_expr LT next_expr { $$ = new_node(NODE_LT, $1, $3); }
              | next_expr GT next_expr { $$ = new_node(NODE_GT, $1, $3); }
              | next_expr LE next_expr { $$ = new_node(NODE_LE, $1, $3); }
              | next_expr GE next_expr { $$ = new_node(NODE_GE, $1, $3); }
              | LCB next_expr_set RCB { $$ = $2; } 
              | next_expr UNION next_expr { $$ = new_node(NODE_UNION, $1, $3); }
              | next_expr SETIN next_expr { $$ = new_node(NODE_SETIN, $1, $3); }
              | CASE n_case_list ESAC { $$ = $2; }
              | SIGMA LB ATOM EQUAL subrange RB next_expr
                { $$ = new_node(NODE_SIGMA, new_node(NODE_EQUAL, $3, $5), $7); }
              ;
next_expr_set : next_expr {$$ = $1;}
              | next_expr_set COMMA next_expr {$$ = new_node(NODE_UNION, $1, $3);}
              ;
n_case_list   : {$$=new_node(NODE_TRUEEXP, NODE_NIL, NODE_NIL);}
              | next_expr COLON next_expr SEMI n_case_list
                {
	          $$ = new_node(NODE_CASE, new_node(NODE_COLON, $1, $3), $5);
	        }
              ;

/* currently not used 
next_expr_list : next_expr {$$ = cons($1,NODE_NIL);}
	         | next_expr_list COMMA next_expr {$$ = cons($3, $1);}
                 ;
*/

/* (Bounded) CTL formulas. Contain CTL operators, but no next variables. */

ctl_expr      : var_id
              | number
              | subrange
              | FALSEEXP
              | TRUEEXP
              | LP ctl_expr RP { $$ = $2; }
	      | ctl_expr IMPLIES ctl_expr { $$ = new_node(NODE_IMPLIES, $1, $3); }
	      | ctl_expr IFF ctl_expr { $$ = new_node(NODE_IFF, $1, $3); }
	      | ctl_expr OR ctl_expr { $$ = new_node(NODE_OR, $1, $3); }
	      | ctl_expr XOR ctl_expr { $$ = new_node(NODE_XOR, $1, $3); }
	      | ctl_expr XNOR ctl_expr { $$ = new_node(NODE_XNOR, $1, $3); }
	      | ctl_expr AND ctl_expr { $$ = new_node(NODE_AND, $1, $3); }
	      | NOT ctl_expr { $$ = new_node(NODE_NOT, $2, NODE_NIL); }
              | ctl_expr PLUS ctl_expr { $$ = new_node(NODE_PLUS, $1, $3); }
              | ctl_expr MINUS ctl_expr { $$ = new_node(NODE_MINUS, $1, $3); }
              | ctl_expr TIMES ctl_expr { $$ = new_node(NODE_TIMES, $1, $3); }
              | ctl_expr DIVIDE ctl_expr { $$ = new_node(NODE_DIVIDE, $1, $3); }
              | ctl_expr MOD ctl_expr { $$ = new_node(NODE_MOD, $1, $3); }
              | ctl_expr EQUAL ctl_expr { $$ = new_node(NODE_EQUAL, $1, $3); }
              | ctl_expr NOTEQUAL ctl_expr { $$ = new_node(NODE_NOTEQUAL, $1, $3); }
              | ctl_expr LT ctl_expr { $$ = new_node(NODE_LT, $1, $3); }
              | ctl_expr GT ctl_expr { $$ = new_node(NODE_GT, $1, $3); }
              | ctl_expr LE ctl_expr { $$ = new_node(NODE_LE, $1, $3); }
              | ctl_expr GE ctl_expr { $$ = new_node(NODE_GE, $1, $3); }
              | LCB ctl_expr_set RCB { $$ = $2; } 
              | ctl_expr UNION ctl_expr { $$ = new_node(NODE_UNION, $1, $3); }
              | ctl_expr SETIN ctl_expr { $$ = new_node(NODE_SETIN, $1, $3); }
              | CASE ctl_case_list ESAC { $$ = $2; }
/*            | SIGMA LB ATOM EQUAL subrange RB ctl_expr */
/*                { $$ = new_node(NODE_SIGMA, new_node(NODE_EQUAL, $3, $5), $7); } */
              | EX ctl_expr { $$ = new_node(NODE_EX, $2, NODE_NIL); }
              | AX ctl_expr { $$ = new_node(NODE_AX, $2, NODE_NIL); }
              | EF ctl_expr { $$ = new_node(NODE_EF, $2, NODE_NIL); }
              | AF ctl_expr { $$ = new_node(NODE_AF, $2, NODE_NIL); }
              | EG ctl_expr { $$ = new_node(NODE_EG, $2, NODE_NIL); }
              | AG ctl_expr { $$ = new_node(NODE_AG, $2, NODE_NIL); }
	      | AA LB ctl_expr UNTIL ctl_expr RB { $$ = new_node(NODE_AU, $3, $5); }
	      | EE LB ctl_expr UNTIL ctl_expr RB { $$ = new_node(NODE_EU, $3, $5); }
	      | AA LB ctl_expr BUNTIL subrange ctl_expr RB
                       { $$ = new_node(NODE_ABU, new_node(NODE_AU, $3, $6), $5); }
	      | EE LB ctl_expr BUNTIL subrange ctl_expr RB
                       { $$ = new_node(NODE_EBU, new_node(NODE_EU, $3, $6), $5); }
              | EBF subrange ctl_expr { $$ = new_node(NODE_EBF, $3, $2); }
              | ABF subrange ctl_expr { $$ = new_node(NODE_ABF, $3, $2); }
              | EBG subrange ctl_expr { $$ = new_node(NODE_EBG, $3, $2); }
              | ABG subrange ctl_expr { $$ = new_node(NODE_ABG, $3, $2); }
	      ;
ctl_expr_set  : ctl_expr {$$ = $1;}
              | ctl_expr_set COMMA ctl_expr {$$ = new_node(NODE_UNION, $1, $3);}
              ;
ctl_case_list : {$$=new_node(NODE_TRUEEXP, NODE_NIL, NODE_NIL);}
              | ctl_expr COLON ctl_expr SEMI ctl_case_list
                {
	          $$ = new_node(NODE_CASE, new_node(NODE_COLON, $1, $3), $5);
	        }
              ;

/* LTL grammar */
ltl_expr      : ltl_orexpr {$$ = $1;}
              | ltl_orexpr IMPLIES ltl_expr {$$ = new_node(NODE_IMPLIES, $1, $3);}
              | ltl_orexpr IFF ltl_orexpr {$$ = new_node(NODE_IFF, $1, $3);}
              | ltl_orexpr XOR ltl_orexpr {$$ = new_node(NODE_XOR, $1, $3);}
              | ltl_orexpr XNOR ltl_orexpr {$$ = new_node(NODE_XNOR, $1, $3);}
              ;
ltl_orexpr    : ltl_andexpr {$$ = $1;}
              | ltl_orexpr OR ltl_andexpr {$$ = new_node(NODE_OR, $1, $3);}
              ;
ltl_andexpr   : ltl_binary_expr {$$ = $1;}
              | ltl_andexpr AND ltl_binary_expr {$$ = new_node(NODE_AND, $1, $3);}
              ;
ltl_binary_expr : ltl_atomexpr {$$ = $1;}
              | ltl_binary_expr UNTIL ltl_atomexpr {$$ = new_node(NODE_UNTIL, $1, $3);}
              | ltl_binary_expr SINCE ltl_atomexpr {$$ = new_node(NODE_SINCE, $1, $3);}
              | ltl_binary_expr RELEASES ltl_atomexpr 
                {$$ = new_node(NODE_NOT, new_node(NODE_UNTIL, new_node(NODE_NOT, $1, NODE_NIL), 
                new_node(NODE_NOT, $3, NODE_NIL)), NODE_NIL);}
              | ltl_binary_expr TRIGGERED ltl_atomexpr 
                {$$ = new_node(NODE_NOT, new_node(NODE_SINCE, new_node(NODE_NOT, $1, NODE_NIL), 
                new_node(NODE_NOT, $3, NODE_NIL)), NODE_NIL);}
              ;
ltl_atomexpr  : NOT ltl_atomexpr {$$ = new_node(NODE_NOT, $2, NODE_NIL);}
              | OP_NEXT       ltl_atomexpr {$$ = new_node(NODE_OP_NEXT, $2, NODE_NIL);} 
              | OP_PREC       ltl_atomexpr {$$ = new_node(NODE_OP_PREC, $2, NODE_NIL);} 
              | OP_NOTPRECNOT ltl_atomexpr {$$ = new_node(NODE_OP_NOTPRECNOT, $2, NODE_NIL);}               
              | OP_GLOBAL     ltl_atomexpr {$$ = new_node(NODE_OP_GLOBAL, $2, NODE_NIL);}
              | OP_HISTORICAL ltl_atomexpr {$$ = new_node(NODE_OP_HISTORICAL, $2, NODE_NIL);}
              | OP_FUTURE     ltl_atomexpr {$$ = new_node(NODE_OP_FUTURE, $2, NODE_NIL);}
              | OP_ONCE       ltl_atomexpr {$$ = new_node(NODE_OP_ONCE, $2, NODE_NIL);}
              | LP ltl_expr RP {$$ = $2;}
              | CASE ltl_case_expr ESAC {$$ = $2;}
              | ltl_simple_expr
              ;

ltl_case_expr : {$$=new_node(NODE_TRUEEXP, NODE_NIL, NODE_NIL);}
              | ltl_expr COLON ltl_expr SEMI ltl_case_expr
                {
	          $$ = new_node(NODE_CASE, new_node(NODE_COLON, $1, $3), $5);
	        }
              ;

ltl_simple_expr : var_id
                | number
                | subrange
                | FALSEEXP
                | TRUEEXP
                | LP ltl_simple_expr RP {$$ = $2;}
                | ltl_simple_expr PLUS ltl_simple_expr {$$ = new_node(NODE_PLUS, $1, $3);}
                | ltl_simple_expr MINUS ltl_simple_expr {$$ = new_node(NODE_MINUS, $1, $3);}
                | ltl_simple_expr TIMES ltl_simple_expr {$$ = new_node(NODE_TIMES, $1, $3);}
                | ltl_simple_expr DIVIDE ltl_simple_expr {$$ = new_node(NODE_DIVIDE, $1, $3);}
                | ltl_simple_expr MOD ltl_simple_expr {$$ = new_node(NODE_MOD, $1, $3);}
                | ltl_simple_expr EQUAL ltl_simple_expr {$$ = new_node(NODE_EQUAL, $1, $3);}
                | ltl_simple_expr NOTEQUAL ltl_simple_expr {$$ = new_node(NODE_NOTEQUAL, $1, $3);}
                | ltl_simple_expr LT ltl_simple_expr {$$ = new_node(NODE_LT, $1, $3);}
                | ltl_simple_expr GT ltl_simple_expr {$$ = new_node(NODE_GT, $1, $3);}
                | ltl_simple_expr LE ltl_simple_expr {$$ = new_node(NODE_LE, $1, $3);}
                | ltl_simple_expr GE ltl_simple_expr {$$ = new_node(NODE_GE, $1, $3);}
                | LCB ltl_expr_set RCB { $$ = $2; } 
                | ltl_simple_expr UNION ltl_simple_expr { $$ = new_node(NODE_UNION, $1, $3);}
                | ltl_simple_expr SETIN ltl_simple_expr { $$ = new_node(NODE_SETIN, $1, $3);}
                ;
ltl_expr_set    : ltl_expr {$$ = $1;}
                | ltl_expr_set COMMA ltl_expr {$$ = new_node(NODE_UNION, $1, $3);}
                ;
/*
   Only for homomorphism checking (currently disabled).
 */
implements    : IMPLEMENTS ATOM {$$ = new_node(NODE_IMPLEMENTS, $2, NODE_NIL);}
              ;
input         : INPUT var_list {$$ = new_node(NODE_INPUT, $2, NODE_NIL);}
              ;
output        : OUTPUT term_list SEMI {$$ = new_node(NODE_OUTPUT, $2, NODE_NIL);}
              ;
term_list     : var_id {$$ = cons($1, NODE_NIL);}
              | term_list COMMA var_id {$$ = cons($3, $1);}
              ;



%%

#if 0
/* Additional source code */
void yyerror(char *s)
{
    extern options_ptr options;
    extern char yytext[];
    
    start_parsing_err();
    fprintf(stderr, "at token \"%s\": %s\n", yytext, s);
    if (opt_batch(options)) finish_parsing_err();
}

int yywrap()
{
  return(1);
}

#endif
