/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                           */
/*                  This file is part of the program                         */
/*        SMV2LP -- Converting SMV to Integer Linear Programs                */
/*                                                                           */
/*    Copyright (C) 2002-2005 Tobias Achterberg                              */
/*                            Thorsten Koch                                  */
/*                            Robi Malik                                     */
/*                                                                           */
/*                  2002-2005 Konrad-Zuse-Zentrum                            */
/*                            fuer Informationstechnik Berlin                */
/*                                                                           */
/*  SMV2LP is distributed under the terms of the ZIB Academic License.       */
/*                                                                           */
/*  You should have received a copy of the ZIB Academic License              */
/*  along with SMV2LP; see the file COPYING.                                 */
/*                                                                           */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <assert.h>

#include "message.h"
#include "memory.h"
#include "node.h"

Node* node_new(
   NodeType  type,
   Node*     left,
   Node*     right,
   int       lineno,
   int       column)
{
   Node* node;

   assert(type >= NODE_ERROR && type < NODE__MAXVAL);

   allocMemory(&node);
   if( NULL != node )
   {
      node->type   = type;
      node->left   = left;
      node->right  = right;
      node->lineno = lineno;
      node->column = column;
   }
   return node;
}

Node* node_new_bool(
   NodeType  type,
   Bool      value,
   Node*     left,
   Node*     right,
   int       lineno,
   int       column)
{
   Node* node;

   if( NULL != (node = node_new(type, left, right, lineno, column)) )
      node->value.bool = value;

   return node;
}

Node* node_new_numb(
   NodeType  type,
   Longint   value,
   Node*     left,
   Node*     right,
   int       lineno,
   int       column)
{
   Node* node;

   if( NULL != (node = node_new(type, left, right, lineno, column)) )
      node->value.numb = value;

   return node;
}

Node* node_new_strg(
   NodeType    type,
   const char* value,
   Node*       left,
   Node*       right,
   int         lineno,
   int         column)
{
   Node* node;

   assert(type == NODE_ATOM);
   if( NULL != (node = node_new(type, left, right, lineno, column)) )
   {
      duplicateMemoryArray(&node->value.strg, value, strlen(value)+1);
   }

   return node;
}

void node_free(Node** node)
{
   assert(node != NULL);
   if( *node != NULL )
   {
      node_free(&((*node)->left));
      node_free(&((*node)->right));
      if( (*node)->type == NODE_ATOM )
         freeMemoryArray(&(*node)->value.strg);
      freeMemory(node);
      *node = NULL;
   }
}

void node_display(
   const Node*  node)
{
   if( node == NULL )
      return;

   printf("[%3d:%2d]", node->lineno, node->column);

   switch(node->type)
   {
   case NODE_ERROR: 
      printf("ERROR\n");
      break;
   case NODE_FALSEEXP: 
      printf("FALSEEXP\n");
      break;
   case NODE_TRUEEXP: 
      printf("TRUEEXP\n");
      break;
   case NODE_NUMBER: 
      printf("NUMBER (%lld)\n", node->value.numb);
      break;
   case NODE_ATOM: 
      printf("ATOM (%s)\n", node->value.strg);
      break;
   case NODE_DOT: 
      printf("DOT\n");
      break;
   case NODE_GOTO: 
      printf("GOTO\n");
      break;
   case NODE_LET: 
      printf("LET\n");
      break;
   case NODE_EVAL: 
      printf("EVAL\n");
      break;
   case NODE_INIT: 
      printf("INIT\n");
      break;
   case NODE_FAILNESS: 
      printf("FAILNESS\n");
      break;
   case NODE_TRANS: 
      printf("TRANS\n");
      break;
   case NODE_CONSTRAINT: 
      printf("CONSTRAINT\n");
      break;
   case NODE_SIMPWFF: 
      printf("SIMPWFF\n");
      break;
   case NODE_CONTEXT: 
      printf("CONTEXT\n");
      break;
   case NODE_CTLWFF: 
      printf("CTLWFF\n");
      break;
   case NODE_LTWFF: 
      printf("LTWFF\n");
      break;
   case NODE_EQDEF: 
      printf("EQDEF\n");
      break;
   case NODE_CONS: 
      printf("CONS\n");
      break;
   case NODE_VAR: 
      printf("VAR\n");
      break;
   case NODE_IVAR:      
      printf("IVAR\n");
      break;
   case NODE_COLON: 
      printf("COLON\n");
      break;
   case NODE_BOOLEAN: 
      printf("BOOLEAN\n");
      break;
   case NODE_SCALAR: 
      printf("SCALAR\n");
      break;
   case NODE_ARRAY: 
      printf("ARRAY\n");
      break;
   case NODE_PROCESS: 
      printf("PROCESS\n");
      break;
   case NODE_TWODOTS: 
      printf("TWODOTS\n");
      break;
   case NODE_UMINUS: 
      printf("UMINUS\n");
      break;
   case NODE_MODTYPE: 
      printf("MODTYPE\n");
      break;
   case NODE_ASSIGN: 
      printf("ASSIGN\n");
      break;
   case NODE_AND: 
      printf("AND\n");
      break;
   case NODE_SMALLINIT: 
      printf("SMALLINIT\n");
      break;
   case NODE_NEXT: 
      printf("NEXT\n");
      break;
   case NODE_INVAR: 
      printf("INVAR\n");
      break;
   case NODE_DEFINE:
      printf("DEFINE\n");
      break;
   case NODE_JUSTICE: 
      printf("JUSTICE\n");
      break;
   case NODE_COMPASSION: 
      printf("COMPASSION\n");
      break;
   case NODE_SPEC: 
      printf("SPEC\n");
      break;
   case NODE_LTLSPEC: 
      printf("LTLSPEC\n");
      break;
   case NODE_INVARSPEC: 
      printf("INVARSPEC\n");
      break;
   case NODE_COMPUTE: 
      printf("COMPUTE\n");
      break;
   case NODE_MINU: 
      printf("MINU\n");
      break;
   case NODE_MAXU: 
      printf("MAXU\n");
      break;
   case NODE_SELF: 
      printf("SELF\n");
      break;
   case NODE_IMPLIES: 
      printf("IMPLIES\n");
      break;
   case NODE_IFF: 
      printf("IFF\n");
      break;
   case NODE_OR: 
      printf("OR\n");
      break;
   case NODE_XOR: 
      printf("XOR\n");
      break;
   case NODE_XNOR: 
      printf("XNOR\n");
      break;
   case NODE_NOT: 
      printf("NOT\n");
      break;
   case NODE_PLUS: 
      printf("PLUS\n");
      break;
   case NODE_MINUS: 
      printf("MINUS\n");
      break;
   case NODE_TIMES: 
      printf("TIMES\n");
      break;
   case NODE_IMPLEMENTS: 
      printf("IMPLEMENTS\n");
      break;
   case NODE_INPUT: 
      printf("INPUT\n");
      break;
   case NODE_OUTPUT: 
      printf("OUTPUT\n");
      break;
   case NODE_FAIRNESS: 
      printf("FAIRNESS\n");
      break;
   case NODE_DIVIDE: 
      printf("DIVIDE\n");
      break;
   case NODE_MOD: 
      printf("MOD\n");
      break;
   case NODE_EQUAL: 
      printf("EQUAL\n");
      break;
   case NODE_NOTEQUAL: 
      printf("NOTEQUAL\n");
      break;
   case NODE_LT: 
      printf("LT\n");
      break;
   case NODE_GT: 
      printf("GT\n");
      break;
   case NODE_LE: 
      printf("LE\n");
      break;
   case NODE_GE: 
      printf("GE\n");
      break;
   case NODE_UNION: 
      printf("UNION\n");
      break;
   case NODE_SETIN: 
      printf("SETIN\n");
      break;
   case NODE_CASE: 
      printf("CASE\n");
      break;
   case NODE_EX: 
      printf("EX\n");
      break;
   case NODE_AX: 
      printf("AX\n");
      break;
   case NODE_EF: 
      printf("EF\n");
      break;
   case NODE_AF: 
      printf("AF\n");
      break;
   case NODE_EG: 
      printf("EG\n");
      break;
   case NODE_AG: 
      printf("AG\n");
      break;
   case NODE_AU: 
      printf("AU\n");
      break;
   case NODE_EU: 
      printf("EU\n");
      break;
   case NODE_EBF: 
      printf("EBF\n");
      break;
   case NODE_ABF: 
      printf("ABF\n");
      break;
   case NODE_EBG: 
      printf("EBG\n");
      break;
   case NODE_ABG: 
      printf("ABG\n");
      break;
   case NODE_UNTIL: 
      printf("UNTIL\n");
      break;
   case NODE_SINCE: 
      printf("SINCE\n");
      break;
   case NODE_OP_NEXT: 
      printf("OP_NEXT\n");
      break;
   case NODE_OP_PREC: 
      printf("OP_PREC\n");
      break;
   case NODE_OP_NOTPRECNOT: 
      printf("OP_NOTPRECNOT\n");
      break;
   case NODE_OP_GLOBAL: 
      printf("OP_GLOBAL\n");
      break;
   case NODE_OP_HISTORICAL: 
      printf("OP_HISTORICAL\n");
      break;
   case NODE_OP_FUTURE: 
      printf("OP_FUTURE\n");
      break;
   case NODE_OP_ONCE: 
      printf("OP_ONCE\n");
      break;
   case NODE_LTLWFF: 
      printf("LTLWFF\n");
      break;
   case NODE_COMPWFF: 
      printf("COMPWFF\n");
      break;
   case NODE_MODULE: 
      printf("MODULE\n");
      break;
   case NODE_ABU: 
      printf("ABU\n");
      break;
   case NODE_EBU:
      printf("EBU\n");
      break;
   case NODE_ISA:
      printf("ISA\n");
      break;
   case NODE_SIGMA:
      printf("SIGMA\n");
      break;
   default:
      errorMessage("invalid node type");
      printf(" -> type=%d\n", node->type);
      abort();
   }
}
static void tree_display_rec(
   const Node* node,
   int         depth)
{
   int i;
   
   if(node == NULL)
      return;

   for( i = 0; i < depth; ++i )
      printf(" ");

   node_display(node);
   tree_display_rec(node->left, depth+1);
   tree_display_rec(node->right, depth+1);
}

void tree_display(
   const Node* node)
{
   tree_display_rec(node, 0);
}
